from policyengine_us.model_api import *


class new_clean_vehicle_battery_capacity(Variable):
    value_type = float
    entity = TaxUnit
    label = "Battery capacity of a purchased new clean vehicle"
    documentation = "In kilowatt-hours (kWh)"
    unit = "kWh"
    definition_period = YEAR
    quantity_type = STOCK



from policyengine_us.model_api import *


class ca_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "CA income tax"
    defined_for = StateCode.CA
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/forms/Search/Home/Confirmation"

    adds = [
        "ca_income_tax_before_refundable_credits",
        "ca_amt",
        "ca_mental_health_services_tax",
        "ca_use_tax",
    ]
    subtracts = ["ca_refundable_credits"]



from policyengine_us.model_api import *


class ut_at_home_parent_credit_agi_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Utah at-home parent credit based on adjusted gross income"
    definition_period = YEAR
    defined_for = StateCode.UT
    reference = (
        "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S1005.html",
        "https://www.taxformfinder.org/forms/2021/2021-utah-tc-40-full-packet.pdf#page=23",
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ut.tax.income.credits.at_home_parent

        return tax_unit("adjusted_gross_income", period) <= p.max_agi



from policyengine_us.model_api import *


class me_property_tax_fairness_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the maine property tax fairness credit"
    definition_period = YEAR
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        # Separate filers are ineligible
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        separate = filing_status == filing_statuses.SEPARATE
        return ~separate



from policyengine_us.model_api import *


class nh_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Hampshire income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NH

    def formula(tax_unit, period, parameters):
        income = max_(0, tax_unit("nh_taxable_income", period))
        p = parameters(period).gov.states.nh.tax.income
        if p.in_effect:
            return income * p.rate
        return 0



from policyengine_us.model_api import *


class mt_capital_gains_tax_applicable_threshold_indiv(Variable):
    value_type = float
    entity = Person
    label = "Montana applicable threshold for the capital gains tax when married couples file separately"
    unit = USD
    definition_period = YEAR
    reference = "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/12/Form_2_2023_Instructions.pdf#page=6"  # Net Long-Term Capital Gains Tax Table
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.main.capital_gains
        capital_gains = person("long_term_capital_gains", period)
        taxable_income = person("mt_taxable_income_indiv", period)
        filing_status = person.tax_unit(
            "state_filing_status_if_married_filing_separately_on_same_return",
            period,
        )
        status = filing_status.possible_values
        non_qualified_income = max_(taxable_income - capital_gains, 0)
        rate_threshold = select(
            [
                filing_status == status.SINGLE,
                filing_status == status.SEPARATE,
                filing_status == status.HEAD_OF_HOUSEHOLD,
                filing_status == status.SURVIVING_SPOUSE,
            ],
            [
                p.rates.single.thresholds[-1],
                p.rates.separate.thresholds[-1],
                p.rates.head_of_household.thresholds[-1],
                p.rates.surviving_spouse.thresholds[-1],
            ],
        )
        return max_(rate_threshold - non_qualified_income, 0)



from policyengine_us.model_api import *


class co_ccap_parent_fee(Variable):
    value_type = float
    entity = SPMUnit
    label = "Colorado Child Care Assistance Program parent fee"
    reference = (
        "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=41",
        "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=62",
    )
    unit = USD
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        year = period.start.year
        if period.start.month >= 10:
            instant_str = f"{year}-10-01"
        else:
            instant_str = f"{year - 1}-10-01"
        p = parameters(instant_str).gov.states.co.ccap
        eligible_children = spm_unit("co_ccap_eligible_children", period)
        base_parent_fee = spm_unit("co_ccap_base_parent_fee", period)
        add_on_parent_fee = spm_unit("co_ccap_add_on_parent_fee", period)

        # Childcare-hours-per-day also affects parent fee.
        # Since each child may need different hours of childcare per day, we
        # have to calculate parent fee one by one and sum them up.
        person = spm_unit.members
        child_age_eligible = person("co_ccap_child_eligible", period)
        childcare_hours_per_day = person(
            "childcare_hours_per_day", period.this_year
        )
        rate = p.parent_fee_rate_by_child_care_hours.calc(
            childcare_hours_per_day, right=True
        )
        childs_parent_fee = spm_unit.project(
            base_parent_fee + add_on_parent_fee
        )
        unrounded_non_discounted_fee = spm_unit.sum(
            child_age_eligible * childs_parent_fee * rate
        )

        non_discounted_fee = np.round(unrounded_non_discounted_fee, 2)
        # Rating of child care facilities also affects parent fee.
        # For households utilizing multiple child care providers, only one
        # child care provider is required to be eligible for the reduced
        # parent fee to apply.
        rating = person(
            "co_quality_rating_of_child_care_facility", period.this_year
        )
        maximum_rating = spm_unit.max(rating * child_age_eligible)
        discounted_rate = p.is_quality_rating_discounted.calc(maximum_rating)
        unrounded = non_discounted_fee * discounted_rate
        return np.round(unrounded, 2)



from policyengine_us.model_api import *


class md_two_income_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD two-income married couple subtraction from AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.marylandtaxes.gov/forms/21_forms/Resident_Booklet.pdf#page=16"
        "https://govt.westlaw.com/mdc/Document/NF93A7BD2E6C811ECA065A3F5EAA0E5C9?viewType=FullText&originationContext=documenttoc&transitionType=CategoryPageItem&contextData=(sc.Default)"
    )
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        is_joint = filing_status == filing_status.possible_values.JOINT
        person = tax_unit.members
        # compute head and spouse US AGI portions using irs_gross_income shares
        us_agi = tax_unit("adjusted_gross_income", period)
        gross_income = person("irs_gross_income", period)
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        head_gross_income = tax_unit.sum(where(is_head, gross_income, 0))
        couple_gross_income = tax_unit.sum(
            where(is_head | is_spouse, gross_income, 0)
        )
        # Compute the head's share of the couple's cross income.
        # Use a mask rather than where to avoid a divide-by-zero warning. Default to one.
        head_frac = np.ones_like(couple_gross_income)
        mask = couple_gross_income > 0
        head_frac[mask] = head_gross_income[mask] / couple_gross_income[mask]
        head_us_agi = head_frac * us_agi
        spouse_us_agi = (1 - head_frac) * us_agi

        # compute head and spouse MD AGI additions using ad hoc rule
        total_additions = tax_unit("md_total_additions", period)
        head_adds = 0.5 * total_additions
        spouse_adds = 0.5 * total_additions

        # sum head and spouse MD AGI subtractions (other than two-income)
        head_subs = 0
        spouse_subs = 0
        p = parameters(period).gov.states.md.tax.income.agi.subtractions
        for sub in p.sources:
            if sub == "md_two_income_subtraction":
                continue
            if sub in ["md_pension_subtraction", "md_socsec_subtraction"]:
                # person-level subtractions
                ind_sub = person(sub + "_amount", period)
                head_subs += tax_unit.sum(is_head * ind_sub)
                spouse_subs += tax_unit.sum(is_spouse * ind_sub)
            else:
                # taxunit-level subtractions
                unit_sub = tax_unit(sub, period)
                head_subs += 0.5 * unit_sub
                spouse_subs += 0.5 * unit_sub

        # compute MD two-income subtraction
        min_agi_adds_subs = min_(
            head_us_agi + head_adds - head_subs,
            spouse_us_agi + spouse_adds - spouse_subs,
        )
        capped_min_agi_adds_subs = min_(
            p.max_two_income_subtraction,
            min_agi_adds_subs,
        )
        return is_joint * max_(0, capped_min_agi_adds_subs)



from policyengine_us.model_api import *


class ar_standard_deduction_indiv(Variable):
    value_type = float
    entity = Person
    label = "Arkansas standard deduction when married couples are filing separately"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=14"
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        filing_status = person.tax_unit(
            "state_filing_status_if_married_filing_separately_on_same_return",
            period,
        )
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        p = parameters(period).gov.states.ar.tax.income.deductions
        return head_or_spouse * p.standard[filing_status]



from policyengine_us.model_api import *


class ctc_limiting_tax_liability(Variable):
    value_type = float
    entity = TaxUnit
    label = "CTC-limiting tax liability"
    unit = USD
    documentation = "The tax liability used to determine the maximum amount of the non-refundable CTC. Excludes SALT from all calculations (this is an inaccuracy required to avoid circular dependencies)."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        simulation = tax_unit.simulation
        no_salt_branch = simulation.get_branch("no_salt")
        no_salt_branch.set_input(
            "salt_deduction", period, np.zeros(tax_unit.count)
        )
        tax_liability_before_credits = no_salt_branch.calculate(
            "income_tax_before_credits", period
        )
        non_refundable_credits = parameters(
            period
        ).gov.irs.credits.non_refundable
        non_refundable_credits_ex_ctc = [
            x for x in non_refundable_credits if x != "non_refundable_ctc"
        ]
        total_credits = add(tax_unit, period, non_refundable_credits_ex_ctc)

        return max_(0, tax_liability_before_credits - total_credits)



from policyengine_us.model_api import *
from policyengine_core.simulations import Simulation
from policyengine_us.tools.geography.county_helpers import (
    map_county_string_to_enum,
)
from policyengine_us.variables.household.demographic.geographic.county.county_enum import (
    County,
)
from policyengine_us.data import ZIP_CODE_DATASET
from policyengine_us.tools.geography.county_helpers import (
    load_county_fips_dataset,
)


class county(Variable):
    value_type = Enum
    possible_values = County
    default_value = County.UNKNOWN
    entity = Household
    label = "County"
    definition_period = YEAR

    def formula(household, period, parameters):
        simulation: Simulation = household.simulation

        # First look if county FIPS is provided; if so, map to county name
        county_fips: "pd.Series[str]" | None = household("county_fips", period)

        if not simulation.is_over_dataset and county_fips.all():
            COUNTY_FIPS_DATASET: "pd.DataFrame" = load_county_fips_dataset()

            # Decode FIPS codes
            county_fips_codes = COUNTY_FIPS_DATASET.set_index("county_fips")
            county_name = county_fips_codes.loc[county_fips, "county_name"]
            state_code = county_fips_codes.loc[county_fips, "state"]
            return map_county_string_to_enum(county_name, state_code)

        # Check if zip_code was explicitly provided as input
        # The zip_code variable auto-generates values, so we need to check if it was user input
        input_variables = getattr(simulation, "input_dataset", {})
        if (
            isinstance(input_variables, dict)
            and "zip_code" not in input_variables
        ):
            # No ZIP code was provided by user, use first county in state
            return household("first_county_in_state", period)

        # Attempt to look up from ZIP code (only if explicitly provided)
        try:
            zip_code = household("zip_code", period).astype(int)
            zip_codes = ZIP_CODE_DATASET.set_index("zip_code")
            county_name = zip_codes.county[zip_code]
            state_code = zip_codes.state[zip_code]
            return map_county_string_to_enum(county_name, state_code)
        except:
            # If ZIP code lookup fails, use first county in state as fallback
            return household("first_county_in_state", period)



from policyengine_us.model_api import *


class il_aabd_shelter_allowance(Variable):
    value_type = float
    entity = Person
    label = (
        "Illinois Aid to the Aged, Blind or Disabled (AABD) shelter allowance"
    )
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.IL
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.248",
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.dhs.aabd.payment.shelter_allowance
        rent_expense = person("rent", period)
        renter = rent_expense > 0
        rent_allowance = min_(rent_expense, p.rent)
        # Housing cost = rent + property tax + homeowners insurance + HOA fees.
        homestead_property_cost = (
            person.spm_unit("housing_cost", period) - rent_expense
        )
        homestead_property_allowance = min_(
            homestead_property_cost, p.homestead
        )

        total_allowance = where(
            renter, rent_allowance, homestead_property_allowance
        )
        # Prorate the total shelter_allowance across all household members
        size = person.spm_unit("spm_unit_size", period)
        return total_allowance / size



from policyengine_us.model_api import *


class ma_part_a_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part A taxable income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions"
    defined_for = StateCode.MA

    adds = [
        "ma_part_a_taxable_dividend_income",
        "ma_part_a_taxable_capital_gains_income",
    ]



from policyengine_us.model_api import *


class traditional_ira_contributions(Variable):
    value_type = float
    entity = Person
    label = "Traditional IRA contributions"
    unit = USD
    documentation = (
        "Contributions to traditional Individual Retirement Accounts."
    )
    definition_period = YEAR



from policyengine_us.model_api import *


class co_denver_property_tax_relief_homeowner_eligible(Variable):
    value_type = float
    entity = SPMUnit
    label = "Eligible for the homeowner Denver Property Tax Relief"
    defined_for = "in_denver"
    definition_period = YEAR
    reference = "https://library.municode.com/co/denver/codes/code_of_ordinances?nodeId=TITIIREMUCO_CH53TAMIRE_ARTXIREPRTAASELLCOPROWTE_S53-492DE"  # 53-495 (d)

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.local.co.denver.dhs
        pays_property_taxes = add(spm_unit, period, ["real_estate_taxes"]) > 0
        has_elderly_or_disabled = spm_unit(
            "has_co_denver_dhs_elderly_disabled", period
        )
        count_dependents = add(spm_unit, period, ["tax_unit_dependents"])
        has_dependents = count_dependents > 0
        homeowners_with_elderly_or_disabled_or_dependents = (
            has_elderly_or_disabled | has_dependents
        ) & pays_property_taxes

        income = spm_unit("co_denver_property_tax_relief_income", period)
        ami = spm_unit.household("ami", period)
        moderate_factor = spm_unit("hud_moderate_income_factor", period)
        homeowner_income_limit = (
            ami * moderate_factor * p.property_tax_relief.ami_rate.homeowner
        )

        homeowner_income_eligible = income <= homeowner_income_limit

        return (
            homeowners_with_elderly_or_disabled_or_dependents
            & homeowner_income_eligible
        )



from policyengine_us.model_api import *


class az_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona deductions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AZ
    reference = "https://azdor.gov/sites/default/files/2023-08/FORMS_INDIVIDUAL_2022_140f.pdf"  # Line 43

    def formula(tax_unit, period, parameters):
        itemized_ded = tax_unit("az_itemized_deductions", period)
        standard_ded = tax_unit("az_standard_deduction", period)
        return max_(itemized_ded, standard_ded)



from policyengine_us.model_api import *


class ny_real_property_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY real property tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (e)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        rptc = parameters(
            period
        ).gov.states.ny.tax.income.credits.real_property_tax

        # Age-based eligibility.
        person = tax_unit.members
        age = person("age", period)
        is_dependent = person("is_tax_unit_dependent", period)
        aged = age >= rptc.elderly_age
        meets_age_condition = tax_unit.any(~is_dependent & aged)

        # Real-estate-based phase-in.
        real_estate_tax = add(tax_unit, period, ["real_estate_taxes"])
        rent = add(tax_unit, period, ["rent"])
        equivalent_rent = rent * rptc.rent_tax_equivalent
        real_estate_tax_or_equiv = real_estate_tax + equivalent_rent

        # Comparison to income for maximum credit determination.
        agi = tax_unit("adjusted_gross_income", period)
        income_threshold = agi * rptc.excess_real_property_tax.calc(agi)
        excess_rpt = max_(0, real_estate_tax_or_equiv - income_threshold)

        # Means-tested conditions based on property value (cliff).
        assessed_value = add(tax_unit, period, ["assessed_property_value"])
        meets_value_conditions = (
            assessed_value <= rptc.max_property_value
        ) & (equivalent_rent <= rptc.max_rent)

        meets_income_condition = agi < rptc.max_agi

        eligible = meets_value_conditions & meets_income_condition

        maximum_credit = where(
            meets_age_condition,
            rptc.maximum.elderly.calc(agi),
            rptc.maximum.non_elderly.calc(agi),
        )

        credit_amount = rptc.rate * excess_rpt

        return min_(eligible * credit_amount, maximum_credit)



from policyengine_us.model_api import *


class pell_grant_countable_assets(Variable):
    value_type = float
    entity = Person
    label = "Pell Grant countable assets"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class childcare_attending_days_per_month(Variable):
    value_type = int
    entity = Person
    label = "Childcare attending days per month"
    definition_period = YEAR
    unit = "days"



from policyengine_us.model_api import *


class tax_unit_taxable_social_security(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Taxable Social Security benefits"
    documentation = "Social security (OASDI) benefits included in AGI, including tier 1 railroad retirement benefits."
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/86"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.social_security.taxability
        gross_ss = tax_unit("tax_unit_social_security", period)

        # The legislation directs the usage of an income definition that is
        # a particularly modified AGI, plus half of gross Social Security
        # payments. Per IRC Section 86(b)(1), this fraction is always 0.5
        # and is handled by the combined_income_ss_fraction parameter.

        combined_income = tax_unit(
            "tax_unit_combined_income_for_social_security_taxability", period
        )
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        separate = filing_status == status.SEPARATE
        cohabitating = tax_unit("cohabitating_spouses", period)
        # Cohabitating married couples filing separately receive a base
        # and adjusted base amount of 0
        base_amount = where(
            separate & cohabitating,
            p.threshold.base.separate_cohabitating,
            p.threshold.base.main[filing_status],
        )
        adjusted_base_amount = where(
            separate & cohabitating,
            p.threshold.adjusted_base.separate_cohabitating,
            p.threshold.adjusted_base.main[filing_status],
        )

        under_first_threshold = combined_income < base_amount
        under_second_threshold = combined_income < adjusted_base_amount

        combined_income_excess = tax_unit(
            "tax_unit_ss_combined_income_excess", period
        )
        excess_over_adjusted_base = max_(
            0, combined_income - adjusted_base_amount
        )

        # Tier 1: Between base and adjusted base thresholds
        # Per IRC §86(a)(1), "the amount determined under paragraph (1)"
        # Taxable amount is lesser of:
        # - base.benefit_cap * SS benefits [§86(a)(1)(A)]
        # - base.excess * excess over base [§86(a)(1)(B)]
        amount_under_paragraph_1 = min_(
            p.rate.base.benefit_cap * gross_ss,
            p.rate.base.excess * combined_income_excess,
        )

        # Tier 2: Above adjusted base threshold
        # Per IRC §86(a)(2)(A)(ii), the lesser of:
        # - "the amount determined under paragraph (1)" (calculated above)
        # - "one-half of the difference between the adjusted base amount and the base amount"
        bracket_amount = min_(
            amount_under_paragraph_1,
            p.rate.additional.bracket * (adjusted_base_amount - base_amount),
        )

        # Per IRC §86(a)(2), the lesser of:
        # (A) 85% of excess over adjusted base + bracket amount, or
        # (B) 85% of social security benefits
        amount_if_over_second_threshold = min_(
            p.rate.additional.excess * excess_over_adjusted_base
            + bracket_amount,
            p.rate.additional.benefit_cap * gross_ss,
        )

        return select(
            [
                under_first_threshold,
                under_second_threshold,
            ],
            [
                0,
                amount_under_paragraph_1,
            ],
            default=amount_if_over_second_threshold,
        )



from policyengine_us.model_api import *


class ma_child_care_subsidies(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts child care subsidies"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MA
    adds = ["ma_ccfa"]



from policyengine_us.model_api import *


class ctc_value(Variable):
    value_type = float
    entity = TaxUnit
    label = "CTC value"
    unit = USD
    documentation = "Actual value of the Child Tax Credit"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        ctc = tax_unit("ctc", period)
        p = parameters(period).gov.irs.credits.ctc.refundable
        if not p.fully_refundable:
            phase_in = tax_unit("ctc_phase_in", period)
            return min_(ctc, phase_in)
        return ctc



from policyengine_us.model_api import *


class de_refundable_eitc_if_claimed(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware refundable EITC if claimed"
    unit = USD
    documentation = (
        "Refundable EITC credit reducing DE State income tax page 8."
    )
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf"
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.de.tax.income.credits.eitc
        return p.refundable * federal_eitc



from policyengine_us.model_api import *


class az_increased_excise_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Eligible for Arizona Increased Excise Tax Credit"
    reference = "https://www.azleg.gov/viewdocument/?docName=https://www.azleg.gov/ars/43/01072-01.htm"
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.credits.increased_excise
        agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        max_income = p.income_threshold[filing_status]
        return agi <= max_income



from policyengine_us.model_api import *


class able_contributions(Variable):
    value_type = float
    entity = TaxUnit
    label = "ABLE contributions"
    documentation = (
        "Contributions made to an ABLE account by all members of the tax unit."
    )
    unit = USD
    definition_period = YEAR

    adds = ["able_contributions_person"]



from policyengine_us.model_api import *


class ny_ctc_post_2024_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "New York CTC post-2024 eligibility"
    documentation = "Whether the tax unit is eligible for New York CTC under post-2024 rules"
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c-1)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ny.tax.income.credits.ctc

        # Only eligible if post-2024 rules are in effect
        if not p.post_2024.in_effect:
            return False

        person = tax_unit.members
        age = person("age", period)

        # Check if we have qualifying children for post-2024 rules
        qualifies_for_federal_ctc = person("ctc_qualifying_child", period)
        qualifies = qualifies_for_federal_ctc & (age >= p.minimum_age)

        # Check if any children get a non-zero credit amount
        credit_by_age = p.post_2024.amount.calc(age)
        qualifying_children = qualifies & (credit_by_age > 0)

        total_qualifying_children = tax_unit.sum(qualifying_children)
        return total_qualifying_children > 0



from policyengine_us.model_api import *


class hud_utility_allowance(Variable):
    value_type = float
    entity = Household
    label = "HUD utility allowance"
    unit = USD
    documentation = "Utility allowance for HUD programs"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/cfr/text/24/982.517",
        # LA County, only one for now. As of 2023-07-01.
        "https://www.lacda.org/docs/librariesprovider25/public-documents/utility-allowance/utility-allownce-2022.pdf",
    )
    defined_for = "tenant_pays_utilities"

    def formula(household, period, parameters):
        household_bedrooms = household("bedrooms", period)
        is_sro = household("is_sro", period)
        in_la = household("in_la", period)
        la_amount = select(
            [
                is_sro,
                household_bedrooms == 0,
                household_bedrooms == 1,
                household_bedrooms == 2,
                household_bedrooms == 3,
                household_bedrooms == 4,
                household_bedrooms == 5,
                household_bedrooms == 6,
                household_bedrooms == 7,
            ],
            # For LA County, assume all utilities are included.
            [
                179,  # SRO: 9 + 15 + 4 + 6 + 6 + 12 + 20 + 55 + 32 + 7 + 13
                227,  # 0bd: 9 + 16 + 5 + 8 + 8 + 16 + 27 + 73 + 42 + 10 + 13
                271,  # 1bd: 13 + 22 + 7 + 11 + 11 + 22 + 33 + 82 + 42 + 15 + 13
                318,  # 2bd: 16 + 29 + 9 + 15 + 15 + 29 + 40 + 91 + 42 + 19 + 13
                375,  # 3bd: 20 + 37 + 11 + 18 + 18 + 35 + 48 + 109 + 42 + 24 + 13
                452,  # 4bd: 27 + 50 + 14 + 23 + 25 + 44 + 55 + 127 + 42 + 32 + 13
                517,  # 5bd: 31 + 58 + 16 + 26 + 31 + 51 + 66 + 145 + 42 + 38 + 13
                580,  # 6bd: 37 + 65 + 18 + 29 + 36 + 58 + 75 + 163 + 42 + 44 + 13
                647,  # 7bd: 42 + 74 + 20 + 33 + 42 + 65 + 85 + 181 + 42 + 50 + 13
            ],
            default=719,  # 8bd: 47 + 86 + 24 + 36 + 47 + 73 + 96 + 199 + 42 + 56 + 13
        )
        return in_la * la_amount * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class is_older_child_for_medicaid(Variable):
    value_type = bool
    entity = Person
    label = "Older children"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a#l_1_D"

    def formula(person, period, parameters):
        age = person("age", period)
        ma = parameters(
            period
        ).gov.hhs.medicaid.eligibility.categories.older_child
        income = person("medicaid_income_level", period)
        is_older_child = ma.age_range.calc(age)
        state = person.household("state_code_str", period)
        income_limit = ma.income_limit[state]
        return is_older_child & (income < income_limit)



from policyengine_us.model_api import *


class property_sales_net_capital_gain(Variable):
    value_type = float
    entity = TaxUnit
    label = "Net capital gains from sale or exchange of property"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class income_tax_capped_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "non-refundable tax credits"
    unit = USD
    documentation = "Capped value of non-refundable tax credits"
    definition_period = YEAR
    adds = ["income_tax_non_refundable_credits"]
    subtracts = ["income_tax_unavailable_non_refundable_credits"]


class income_tax_unavailable_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "unavailable non-refundable tax credits"
    unit = USD
    documentation = "Total value of non-refundable tax credits that were not available to the filer due to having too low income tax."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return -min_(
            tax_unit("income_tax_before_credits", period),
            tax_unit("income_tax_non_refundable_credits", period),
        ) + tax_unit("income_tax_non_refundable_credits", period)



from policyengine_us.model_api import *


class is_ssi_blind_or_disabled_working_student_exclusion_eligible(Variable):
    value_type = float
    entity = Person
    label = "Eligible for SSI blind or disabled working student earned income exclusion"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/20/416.1112#c_3"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.ssa.ssi.income.exclusions.blind_or_disabled_working_student
        is_blind = person("is_blind", period)
        is_disabled = person("is_disabled", period)
        demographic_eligible = is_blind | is_disabled
        under_age_limit = person("age", period) < p.age_limit
        eligible_student = under_age_limit & person(
            "is_full_time_student", period
        )
        return eligible_student & demographic_eligible



from policyengine_us.model_api import *


class capped_energy_efficient_window_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Capped energy efficient window credit"
    documentation = (
        "Capped credit on energy-efficient exterior window and skylights"
    )
    unit = USD
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=339"

    def formula(tax_unit, period, parameters):
        expenditure = tax_unit("energy_efficient_window_expenditures", period)
        p = parameters(
            period
        ).gov.irs.credits.energy_efficient_home_improvement
        rate = p.rates.improvements
        uncapped = expenditure * rate
        # First cap by year, then lifetime.
        capped_annual = min_(uncapped, p.cap.annual.window)
        prior_credits = tax_unit(
            "prior_energy_efficient_window_credits", period
        )
        return min_(capped_annual, p.cap.lifetime.window - prior_credits)



from policyengine_us.model_api import *


class md_tax_unit_earned_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD tax unit earned income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD

    formula = sum_among_non_dependents("earned_income")



from policyengine_us.model_api import *


class oh_retirement_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio Retirement Income Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://codes.ohio.gov/ohio-revised-code/section-5747.055"
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        retirement_credit = tax_unit(
            "oh_pension_based_retirement_income_credit", period
        )
        lump_sum_credit = tax_unit("oh_lump_sum_retirement_credit", period)
        return max_(retirement_credit, lump_sum_credit)



from policyengine_us.model_api import *


class tx_tanf_eligible_child(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Eligible child for Texas TANF (included in certified group)"
    reference = (
        "https://www.law.cornell.edu/regulations/texas/1-Tex-Admin-Code-SS-372-104",
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-220-tanf",
    )
    defined_for = StateCode.TX

    def formula(person, period, parameters):
        # Child must meet age/demographic requirements
        age_eligible = person("tx_tanf_age_eligible_child", period)

        # And meet inclusion requirements (immigration eligible, not receiving SSI)
        inclusion_requirements = person(
            "tx_tanf_categorically_eligible_person", period
        )

        return age_eligible & inclusion_requirements



from policyengine_us.model_api import *


class estate_tax_credit(Variable):
    value_type = float
    entity = Person
    label = "Estate tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/2010"
    defined_for = "is_deceased"

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.credits.estate
        # A deceased spouse unused exclusion amount can be computed.
        return p.base



from policyengine_us.model_api import *


class ccdf_income_to_smi_ratio(Variable):
    value_type = float
    entity = SPMUnit
    label = "Income to SMI ratio"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        income = spm_unit("ccdf_income", period)
        smi = spm_unit("hhs_smi", period)
        return income / smi



from policyengine_us.model_api import *


class md_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD income tax before credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        taxable_income = tax_unit("md_taxable_income", period)

        # Calculate regular income tax based on filing status
        p = parameters(period).gov.states.md.tax.income
        regular_income_tax = select(
            [
                filing_status == filing_statuses.SINGLE,
                filing_status == filing_statuses.SEPARATE,
                filing_status == filing_statuses.JOINT,
                filing_status == filing_statuses.HEAD_OF_HOUSEHOLD,
                filing_status == filing_statuses.SURVIVING_SPOUSE,
            ],
            [
                p.rates.single.calc(taxable_income),
                p.rates.separate.calc(taxable_income),
                p.rates.joint.calc(taxable_income),
                p.rates.head_of_household.calc(taxable_income),
                p.rates.surviving_spouse.calc(taxable_income),
            ],
        )

        # Add capital gains surtax if applicable
        if p.capital_gains.surtax_applies:
            capital_gains_surtax = tax_unit("md_capital_gains_surtax", period)
            return regular_income_tax + capital_gains_surtax
        return regular_income_tax



from policyengine_us.model_api import *


class va_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
        "https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2022/760-2022.pdf#page=1"
    )
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        std_ded = tax_unit("va_standard_deduction", period)
        itm_ded = tax_unit("va_itemized_deductions", period)
        itemizes = tax_unit("tax_unit_itemizes", period)
        return where(itemizes, itm_ded, std_ded)



from policyengine_us.model_api import *


class ca_fera_eligible(Variable):
    value_type = bool
    entity = Household
    definition_period = YEAR
    label = "Eligible for California FERA program"
    documentation = "Eligible for California Alternate Rates for Energy"
    reference = (
        "https://www.cpuc.ca.gov/industries-and-topics/electrical-energy/electric-costs/care-fera-program",
        "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=PUC&sectionNum=739.12",
    )
    defined_for = StateCode.CA

    def formula(household, period, parameters):
        # Check not eligible for CARE
        care_eligible = household("ca_care_eligible", period)
        # Check at least 3 people in household
        n = household("household_size", period)
        p = parameters(period).gov.states.ca.cpuc.fera.eligibility
        eligible_household_size = n >= p.minimum_household_size
        # Check income eligibility with respect to percent of the poverty line.
        # Must be above 200% of the poverty line (CARE requirements), but less
        # than or equal to 250% of the poverty line.
        income = household("household_market_income", period)
        ca_care_poverty_line = household("ca_care_poverty_line", period)
        income_limit = ca_care_poverty_line * p.fpl_limit
        income_eligible = income <= income_limit
        return income_eligible & eligible_household_size & ~care_eligible



from policyengine_us.model_api import *


class never_eligible_for_social_security_benefits(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Never eligible for Social Security"
    unit = USD

    def formula(person, period, parameters):
        # Assumption: if person is above age 70 and has no social security benefits, then they are never eligible.
        age = person("age", period)
        social_security = person("social_security", period)
        return (age >= 70) & (social_security == 0)



from policyengine_us.model_api import *


class taxsim_mstat(Variable):
    value_type = int
    entity = TaxUnit
    label = "Marital Status"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        fstatus = filing_status.possible_values
        return select(
            [
                filing_status == fstatus.SINGLE,
                filing_status == fstatus.HEAD_OF_HOUSEHOLD,
                filing_status == fstatus.JOINT,
                filing_status == fstatus.SEPARATE,
                filing_status == fstatus.SURVIVING_SPOUSE,
            ],
            [
                1,
                1,
                2,
                6,
                8,
            ],
        )



from policyengine_us.model_api import *


class spm_unit_count_adults(Variable):
    value_type = int
    entity = SPMUnit
    label = "adults in SPM unit"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        return spm_unit.sum(spm_unit.members("is_adult", period))



from policyengine_us.model_api import *


class az_charitable_contributions_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona charitable contributions credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AZ
    reference = (
        "https://law.justia.com/codes/arizona/2022/title-43/section-43-1088/"
    )

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.credits.charitable_contribution.ceiling
        charitable_contributions = tax_unit(
            "az_charitable_contributions_to_qualifying_charitable_organizations",
            period,
        )
        foster_care_contributions = tax_unit(
            "az_charitable_contributions_to_qualifying_foster_care_organizations",
            period,
        )
        filing_status = tax_unit("filing_status", period)
        capped_charitable_contributions = min_(
            charitable_contributions, p.qualifying_organization[filing_status]
        )
        capped_foster_care_contributions = min_(
            foster_care_contributions, p.qualifying_foster[filing_status]
        )
        return (
            capped_charitable_contributions + capped_foster_care_contributions
        )



from policyengine_us.model_api import *


class vt_renter_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont renter credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/vermont/2022/title-32/chapter-154/section-6066/"  # b
        "https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf#page=35"
        # the formula used in this file is based on the excel sheet provided on the official Vermont government website
        "https://tax.vermont.gov/individuals/renter-credit/calculator-and-credit-amounts"  # link for the excel sheet
    )
    defined_for = StateCode.VT

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.vt.tax.income.credits.renter
        # get tax unit size and county.
        tax_unit_size = tax_unit("tax_unit_size", period)
        county = tax_unit.household("county", period)
        # locate the values by family size and county
        full_credit_income_limit = p.income_limit_ami.thirty_percent[
            tax_unit_size
        ][county]
        partial_credit_income_limit = p.income_limit_ami.fifty_percent[
            tax_unit_size
        ][county]
        fmr = p.fair_market_rent[tax_unit_size][county]
        base_credit_amount = fmr * MONTHS_IN_YEAR * p.fmr_rate

        # Compute what the spreadsheet calls the "percent rebate claimable"
        vt_renter_credit_income = tax_unit("vt_renter_credit_income", period)
        income_diff = partial_credit_income_limit - vt_renter_credit_income
        income_threshold_diff = (
            partial_credit_income_limit - full_credit_income_limit
        )
        percent_rebate_claimable = income_diff / income_threshold_diff
        # If shared residence, reduce by a given fraction.
        shared_rent = tax_unit("rent_is_shared_with_another_tax_unit", period)
        shared_residence_reduction = shared_rent * p.shared_residence_reduction
        # if subsidized, get base credit
        has_housing_assistance = (
            tax_unit.spm_unit("housing_assistance", period) > 0
        )
        rent_amount = add(tax_unit, period, ["rent"])
        base_credit_subsidized = rent_amount * p.fmr_rate

        # income > partial credit income limit, then credit is 0
        # income < full credit income limit, then credit is fair market rent
        # income is between partial credit income limit and full credit income limit:
        ## if no subsidized: (partial credit income limit - income)/(partial credit income limit - full credit income limit) * fair market rent
        ## if subsidized: (partial credit income limit - income)/(partial credit income limit - full credit income limit) * actual pay rent amount * rent rate
        # for all of the above situations, multiply by the share rent rate if the the rental unit was shared
        high_income = vt_renter_credit_income > partial_credit_income_limit
        low_income = vt_renter_credit_income < full_credit_income_limit
        mid_income = ~(high_income | low_income)
        credit_value = select(
            [
                low_income & ~has_housing_assistance,
                low_income & has_housing_assistance,
                mid_income & ~has_housing_assistance,
                mid_income & has_housing_assistance,
            ],
            [
                base_credit_amount,
                base_credit_subsidized,
                percent_rebate_claimable * base_credit_amount,
                percent_rebate_claimable * base_credit_subsidized,
            ],
            default=0,
        )
        unrounded = credit_value * (1 - shared_residence_reduction)
        return np.round(unrounded)



from policyengine_us.model_api import *


class aca_required_contribution_percentage(Variable):
    value_type = float
    entity = TaxUnit
    label = "ACA PTC phase-out rate (i.e., IRS Form 8962 'applicable figure')"
    unit = "/1"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/36B#b_3_A"

    def formula(tax_unit, period, parameters):
        magi_frac = tax_unit("aca_magi_fraction", period)
        p = parameters(period).gov.aca.required_contribution_percentage
        return np.interp(magi_frac, p.thresholds, p.amounts)



from policyengine_us.model_api import *


class ma_liheap_fpg(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts LIHEAP federal poverty guideline"
    definition_period = YEAR
    defined_for = StateCode.MA
    reference = "https://www.mass.gov/doc/fy-2025-heap-income-eligibility-benefit-chart-may-8-2025/download"

    def formula(spm_unit, period, parameters):
        n = spm_unit("spm_unit_size", period.this_year)
        state_group = spm_unit.household("state_group_str", period.this_year)
        year = period.start.year
        instant_str = f"{year - 1}-10-01"
        p_fpg = parameters(instant_str).gov.hhs.fpg
        p1 = p_fpg.first_person[state_group]
        pn = p_fpg.additional_person[state_group]
        return p1 + pn * (n - 1)



from policyengine_us.model_api import *


class new_clean_vehicle_battery_critical_minerals_extracted_in_trading_partner_country(
    Variable
):
    value_type = float
    entity = TaxUnit
    label = "Percent of new clean vehicle's battery critical minerals extracted in a US trading partner country"
    documentation = "Percent of newly purchased new clean vehicle's battery critical minearls (by value) extracted or processed in any country with which the US has a free trade agreement in effect, or recycled in North America"
    unit = "/1"
    definition_period = YEAR



from policyengine_us.model_api import *


class is_tax_unit_head(Variable):
    value_type = bool
    entity = Person
    label = "Head of tax unit"
    definition_period = YEAR

    def formula(person, period, parameters):
        # Only adults can be heads.
        eligible = ~person("is_child", period)
        age = person("age", period)
        tax_unit = person.tax_unit
        return person.get_rank(tax_unit, -age, eligible) == 0



from policyengine_us.model_api import *


class base_part_a_premium(Variable):
    value_type = float
    entity = Person
    label = "Medicare Part A Base Premium"
    unit = USD
    documentation = (
        "Annual Medicare Part A premium. Most people (99%) qualify for "
        "premium-free Part A with 40+ quarters of Medicare-covered employment. "
        "Those with 30-39 quarters pay a reduced premium, and those with "
        "fewer than 30 quarters pay the full premium."
    )
    definition_period = YEAR
    reference = "https://www.cms.gov/newsroom/fact-sheets/2025-medicare-parts-b-premiums-and-deductibles"
    defined_for = "is_medicare_eligible"

    def formula(person, period, parameters):
        p = parameters(period).gov.hhs.medicare.part_a
        quarters_covered = person("medicare_quarters_of_coverage", period)

        premium_free_part_a = person("is_premium_free_part_a", period)
        quarters_above_reduced_premium_threshold = (
            quarters_covered >= p.reduced_premium_quarters_threshold
        )
        premium_amount = where(
            quarters_above_reduced_premium_threshold,
            p.reduced_premium,
            p.full_premium,
        )
        return where(premium_free_part_a, 0, premium_amount) * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class sc_retirement_cap(Variable):
    value_type = float
    entity = Person
    label = "South Carolina retirement income subtraction cap"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.SC

    def formula(person, period, parameters):
        p = parameters(period).gov.states.sc.tax.income.subtractions.retirement
        age = person("age", period)
        return p.cap.calc(age)



from policyengine_us.model_api import *


class mo_federal_income_tax_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Missouri Federal income tax deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-1040%20Instructions_2021.pdf#page=7",
        "https://revisor.mo.gov/main/OneSection.aspx?section=143.171&bid=49937&hl=federal+income+tax+deduction%u2044",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        # Deduct a capped share of federal income tax liability.
        # Ignore certain refundable credits: recovery rebates and EITC.
        # See #1528 for uncertainty around these credits.
        p = parameters(
            period
        ).gov.states.mo.tax.income.deductions.federal_income_tax
        uncapped_federal_income_tax_ignoring_credits = add(
            tax_unit, period, ["income_tax"] + p.ignored_credits
        )
        # Limit to nonnegative tax.
        federal_income_tax_ignoring_credits = max_(
            0, uncapped_federal_income_tax_ignoring_credits
        )
        # Apply rate based on MO AGI.
        tax_unit_mo_agi = add(tax_unit, period, ["mo_adjusted_gross_income"])
        rate = p.rate.calc(tax_unit_mo_agi)
        uncapped = federal_income_tax_ignoring_credits * rate
        # Apply cap based on filing status.
        cap = p.cap[tax_unit("filing_status", period)]
        return min_(uncapped, cap)



from policyengine_us.model_api import *


class mo_income_tax_exempt(Variable):
    value_type = float
    entity = Person
    label = "Missouri income tax exempt"
    unit = USD
    definition_period = YEAR
    reference = "https://revisor.mo.gov/main/OneSection.aspx?section=143.021"
    defined_for = StateCode.MO

    def formula(person, period, parameters):
        taxable_income = person("mo_taxable_income", period)
        p = parameters(period).gov.states.mo.tax.income
        return taxable_income <= p.minimum_taxable_income



from policyengine_us.model_api import *


class taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "IRS taxable income"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        agi = tax_unit("adjusted_gross_income", period)
        exemptions = tax_unit("exemptions", period)
        deductions = tax_unit("taxable_income_deductions", period)
        return max_(0, agi - exemptions - deductions)



from policyengine_us.model_api import *


class additional_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Additional standard deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/63#f"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions.standard
        filing_status = tax_unit("filing_status", period)
        amount_per = p.aged_or_blind.amount[filing_status]
        return amount_per * tax_unit("aged_blind_count", period)



from policyengine_us.model_api import *


class ia_regular_tax_joint(Variable):
    value_type = float
    entity = Person
    label = "Iowa regular tax calculated using income tax rate schedule when married couples file jointly"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=53"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=53"
        "https://tax.iowa.gov/sites/default/files/2023-11/IA1041Inst%2863002%29.pdf#page=4"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        taxable_income = person("ia_taxable_income_joint", period)
        p = parameters(period).gov.states.ia.tax.income.rates
        if p.by_filing_status.active:
            filing_status = person.tax_unit(
                "filing_status",
                period,
            )
            joint = filing_status == filing_status.possible_values.JOINT
            return where(
                joint,
                p.by_filing_status.joint.calc(taxable_income),
                p.by_filing_status.other.calc(taxable_income),
            )
        return p.combined.calc(taxable_income)



from policyengine_us.model_api import *


class ar_taxable_capital_gains_indiv(Variable):
    value_type = float
    entity = Person
    label = "Arkansas taxable capital gains when married filing separately"
    unit = USD
    reference = (
        "https://codes.findlaw.com/ar/title-26-taxation/ar-code-sect-26-51-815.html",
        "https://www.taxformfinder.org/forms/2023/2023-arkansas-form-ar1000d.pdf#page=1",
    )
    definition_period = YEAR
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        # Line 1-3 - long term capital gain or loss
        lt_capital_gains = person("long_term_capital_gains", period)
        # Line 4-6 - short term capital loss
        st_capital_gains = person("short_term_capital_gains", period)
        st_capital_loss = max_(-st_capital_gains, 0)
        # Line 7a - Net capital gain or loss
        net_capital_gain = lt_capital_gains - st_capital_loss
        # Line 7b - capped net capital gain
        p = parameters(
            period
        ).gov.states.ar.tax.income.gross_income.capital_gains
        capped_net_cap_gain = min_(net_capital_gain, p.exempt.cap)
        # Line 8 - Tax rate applied to capital gain
        # 50% exempt if a gain, otherwise entire loss.
        exempt_fraction = where(capped_net_cap_gain > 0, p.exempt.rate, 0)
        taxable_amount = capped_net_cap_gain * (1 - exempt_fraction)
        # Lines 9-11: Arkansas short term capital gain if any.
        stcg_if_any = max_(st_capital_gains, 0)
        # Line 12: Total taxable gain or loss. Loss is capped.
        total_taxable_cap_gain_or_loss = taxable_amount + stcg_if_any
        filing_status = person.tax_unit(
            "state_filing_status_if_married_filing_separately_on_same_return",
            period,
        )
        loss_cap = p.loss_cap[filing_status]
        return max_(-loss_cap, total_taxable_cap_gain_or_loss)



from policyengine_us.model_api import *


class pell_grant_contribution_from_assets(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Pell Grant head contribution from assets"
    definition_period = YEAR

    def formula(person, period, parameters):
        assets = person.tax_unit("pell_grant_head_assets", period)
        simplified = person("pell_grant_simplified_formula_applies", period)
        formula = person("pell_grant_formula", period)
        p = parameters(period).gov.ed.pell_grant.head
        return assets * p.asset_assessment_rate[formula] * ~simplified



from policyengine_us.model_api import *


class under_60_days_postpartum(Variable):
    value_type = bool
    entity = Person
    label = "Under 60 days postpartum"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class ia_alternate_tax_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Iowa alternate tax eligible"
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=53"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=53"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        return filing_status != filing_status.possible_values.SINGLE



from policyengine_us.model_api import *


class wv_low_income_earned_income_exclusion_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the West Virginia low-income earned income exclusion"
    defined_for = StateCode.WV
    definition_period = YEAR
    reference = (
        "https://code.wvlegislature.gov/11-21-10/"
        "https://tax.wv.gov/Documents/TaxForms/2021/it140.pdf#page=20"
        "https://tax.wv.gov/Documents/PIT/2022/PersonalIncomeTaxFormsAndInstructions.2022.pdf#page=27"
    )

    def formula(tax_unit, period, parameters):
        federal_agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)

        p = parameters(
            period
        ).gov.states.wv.tax.income.subtractions.low_income_earned_income

        return federal_agi <= p.income_limit[filing_status]



from policyengine_us.model_api import *


class ca_capi_resources(Variable):
    value_type = float
    unit = USD
    entity = SPMUnit
    label = "California CAPI resources"
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "https://www.cdss.ca.gov/Portals/9/CAPI/CAPI_Regulations-Accessible.pdf"

    adds = ["ssi_countable_resources", "ca_capi_countable_vehicle_value"]



from policyengine_us.model_api import *


class ok_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oklahoma taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ok_agi", period)
        adjustments = tax_unit("ok_adjustments", period)
        exemptions = tax_unit("ok_exemptions", period)
        # From page 10 of the 2021 and 2022 Form 511 Packets:
        #   "If you claimed the standard deduction on your federal return,
        #   you must claim the Oklahoma standard deduction. If you claimed
        #   itemized deductions on your federal return, you must claim
        #   Oklahoma itemized deductions."
        std_ded = tax_unit("ok_standard_deduction", period)
        itm_ded = tax_unit("ok_itemized_deductions", period)
        federal_itemizer = tax_unit("tax_unit_itemizes", period)
        deductions = where(federal_itemizer, itm_ded, std_ded)
        return max_(0, agi - adjustments - exemptions - deductions)



from policyengine_us.model_api import *


class pr_gradual_adjustment_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Puerto Rico gradual adjustment eligibility"
    definition_period = YEAR
    defined_for = StateCode.PR
    reference = "https://hacienda.pr.gov/sites/default/files/individuals_2024_rev._jul_12_24_9-30-24_informative.pdf#page=3"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.territories.pr.tax.income.gradual_adjustment
        net_taxable_income = tax_unit("pr_net_taxable_income", period)

        return net_taxable_income > p.threshold



from policyengine_us.model_api import *


class birth_year(Variable):
    value_type = int
    entity = Person
    label = "Birth year"
    unit = "year"
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)
        return period.start.year - age



from policyengine_us.model_api import *


class ut_taxpayer_credit_phase_out_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah taxpayer credit phase-out income"
    unit = USD
    documentation = (
        "Income that reduces the Utah taxpayer credit. Form TC-40, line 18"
    )
    definition_period = YEAR
    defined_for = StateCode.UT

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        thresholds = parameters(
            period
        ).gov.states.ut.tax.income.credits.taxpayer.phase_out.threshold
        threshold = thresholds[filing_status]
        income = tax_unit("ut_taxable_income", period)
        return max_(income - threshold, 0)



from policyengine_us.model_api import *


class residential_efficiency_electrification_retrofit_energy_savings(Variable):
    value_type = float
    entity = TaxUnit
    label = "Modeled energy system savings from a residential efficiency and electrification retrofit"
    documentation = "In kilowatt hours per month. Do not include savings from projects listed in other electrification and efficiency expenditure categories."
    unit = "kWh/month"
    definition_period = YEAR
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=587"



from policyengine_us.model_api import *


class amt_income(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "AMT taxable income"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/56"

    adds = [
        "taxable_income",
        "amt_excluded_deductions",
        "amt_separate_addition",
        "exemptions",
    ]



from policyengine_us.model_api import *


class non_refundable_american_opportunity_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Non-refundable American Opportunity Credit"
    unit = USD
    documentation = "Value of the non-refundable portion of the American Opportunity Credit"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#i"

    def formula(tax_unit, period, parameters):
        credit_limit = tax_unit(
            "non_refundable_american_opportunity_credit_credit_limit", period
        )
        potential = tax_unit(
            "non_refundable_american_opportunity_credit_potential", period
        )
        return min_(credit_limit, potential)



from policyengine_us.model_api import *


class basic_income_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Basic income eligible"
    unit = USD
    documentation = (
        "Eligible for basic income payments based on adjusted gross income."
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.contrib.ubi_center.basic_income.agi_limit
        if not p.in_effect:
            return True
        agi_limit = p.amount[tax_unit("filing_status", period)]
        return tax_unit("adjusted_gross_income", period) <= agi_limit



from policyengine_us.model_api import *


class ms_retirement_income_exemption(Variable):
    value_type = float
    entity = Person
    label = "Mississippi retirement income exemption"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://law.justia.com/codes/mississippi/title-27/chapter-7/article-1/section-27-7-15/",  # (4)(k)
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=11",  # Line 46
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/2024/80105248.pdf",
    ]
    defined_for = StateCode.MS

    def formula(person, period, parameters):
        excluded_income = add(
            person,
            period,
            [
                "taxable_social_security",
                "taxable_pension_income",
            ],
        )
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return head_or_spouse * excluded_income



from policyengine_us.model_api import *


class de_itemized_deductions_indv(Variable):
    value_type = float
    entity = Person
    label = "Delaware itemized deductions when married couples are filing separately"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenuefiles.delaware.gov/2022/TY22_PIT-RSA_2022-02_PaperInteractive.pdf",  # § 1109
        "https://delcode.delaware.gov/title30/c011/sc02/index.html",
    )
    defined_for = StateCode.DE

    def formula(person, period, parameters):
        unit_deds = person.tax_unit("de_itemized_deductions_unit", period)
        person_agi = person("adjusted_gross_income_person", period)
        total_agi = person.tax_unit.sum(person_agi)

        prorate = np.zeros_like(total_agi)
        mask = total_agi > 0
        prorate[mask] = person_agi[mask] / total_agi[mask]
        # Dependents should always return 0 as their AGI is always
        # attributed to the head of the tax unit in de_agi
        return unit_deds * prorate



from policyengine_us.model_api import *


class la_refundable_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana refundable Child and Dependent Care Credit"
    unit = USD
    definition_period = YEAR
    reference = "http://legis.la.gov/Legis/Law.aspx?d=101769"
    defined_for = StateCode.LA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.la.tax.income.credits.cdcc.refundable
        # determine LA cdcc amount
        # Louisiana matches the potential federal credit
        us_cdcc = tax_unit("cdcc_potential", period)
        us_agi = tax_unit("adjusted_gross_income", period)
        return us_cdcc * p.match.calc(us_agi, right=True)



from policyengine_us.model_api import *


class mt_taxable_social_security(Variable):
    value_type = float
    entity = Person
    label = "Montana taxable social security benefits"
    defined_for = StateCode.MT
    definition_period = YEAR
    reference = "https://mtrevenue.gov/wp-content/uploads/mdocs/form%202%202021.pdf#page=6"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.social_security.amount
        p_irs = parameters(period).gov.irs.social_security.taxability.rate
        # Compute the amount based on the schedule in Montana tax form.
        # line 1 total net SS amount
        social_security = person("social_security", period)
        # Line 2 SS multiplied by the base rate
        social_security_benefits_fraction = (
            social_security * p_irs.base.benefit_cap
        )
        # line 3: irs_gross_income - taxable_social_security
        taxable_ss = person("taxable_social_security", period)
        gross_income = person("irs_gross_income", period)
        reduced_gross_income = max_(gross_income - taxable_ss, 0)
        # line 4: Additions Schedule line 15 - Additions Schedule line 3
        # Interest and mutual fund dividends from state, county, or municipal bonds
        # from other states - not included
        # line 5: tax exempt interest income
        tax_exempt_interest_income = person(
            "tax_exempt_interest_income", period
        )
        # line 6: Sum of line 2, 3, 4, 5
        income_increased_by_ss_and_interest = (
            social_security_benefits_fraction
            + reduced_gross_income
            + tax_exempt_interest_income
        )
        # line 7: Remove the student loans from the above the line deductions
        ald_less_student_loan = person(
            "mt_applicable_ald_deductions", period
        ) - person("student_loan_interest_ald", period)
        # Line 8 - Montana subtractions + ald
        subtractions = person("mt_subtractions", period)
        increased_subtractions = subtractions + ald_less_student_loan
        # line 9: line 6 -line 8, if line 8 >= line 6, return 0
        income_reduced_by_subtractions = max_(
            0, income_increased_by_ss_and_interest - increased_subtractions
        )
        # line 10: get amount based on filing status
        filing_status = person.tax_unit("filing_status", period)
        threshold_amount = p.lower[filing_status]
        # line 11: line 9 - line 10 (income_reduced_by_subtractions - threshold_amount)
        income_reduced_by_subtractions_and_threshold = max_(
            0, income_reduced_by_subtractions - threshold_amount
        )
        # line 12: get amount based on filing status
        amount_lower = p.upper[filing_status] - p.lower[filing_status]
        # lien 13: line 11 - line 12
        minimum_tax_threshold = max_(
            0, income_reduced_by_subtractions_and_threshold - amount_lower
        )
        # line 14 & 15
        capped_reduced_income = min_(
            income_reduced_by_subtractions_and_threshold, amount_lower
        )
        minimum_tax_threshold_fraction = (
            capped_reduced_income * p_irs.base.excess
        )
        # line 16
        smaller_fraction = min_(
            minimum_tax_threshold_fraction, social_security_benefits_fraction
        )
        # line 17 & 18 ## line_13*0.85 + line_16
        adjusted_tax_amount = (
            minimum_tax_threshold * p_irs.additional.benefit_cap
            + smaller_fraction
        )
        # line 19 ## line_1*0.85
        adjusted_taxable_amount = social_security * p_irs.additional.excess
        # line 20
        return min_(adjusted_taxable_amount, adjusted_tax_amount)



from policyengine_us.model_api import *


class is_part_time_college_student(Variable):
    value_type = bool
    entity = Person
    label = "Is a part-time college student"
    documentation = "Enrolled at least half-time but less than full-time in an institution of higher education"
    definition_period = YEAR



from policyengine_us.model_api import *


class la_school_readiness_credit_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Eligible child for the Louisiana school readiness tax credit"
    definition_period = YEAR
    reference = "https://revenue.louisiana.gov/TaxForms/IT540WEB(2022)%20F%20D2.pdf#page=15"
    defined_for = StateCode.LA

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.la.tax.income.credits.school_readiness
        dependent = person("is_tax_unit_dependent", period)
        age = person("age", period)
        age_eligible = age < p.age_threshold

        return dependent & age_eligible



from policyengine_us.model_api import *


class il_hbwd_premium(Variable):
    value_type = float
    entity = Person
    label = "Illinois Health Benefits for Workers with Disabilities monthly premium"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://hfs.illinois.gov/medicalprograms/hbwd/premiums.html",
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-120.510",
    )
    defined_for = "il_hbwd_eligible"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.hfs.hbwd.premium

        earned = person("il_hbwd_countable_earned_income", period)
        unearned = person("il_hbwd_countable_unearned_income", period)

        # Lookup premium components from bracket parameters
        return p.earned.calc(earned) + p.unearned.calc(unearned)



from policyengine_us.model_api import *


class ny_allowable_college_tuition_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York allowable college tuition expenses for the credit and deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.nysenate.gov/legislation/laws/TAX/606"  # (t)(2)(A)
    )
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ny.tax.income.college_tuition
        person = tax_unit.members
        capped = min_(person("qualified_tuition_expenses", period), p.cap)
        return tax_unit.sum(capped)



from policyengine_us.model_api import *


class ms_dependents_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Mississippi qualified and other dependent children exemption"
    reference = "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=5"
    definition_period = YEAR
    defined_for = StateCode.MS

    def formula(tax_unit, period, parameters):
        # Then get the MS Exemptions part of the parameter tree.
        p = parameters(period).gov.states.ms.tax.income.exemptions.dependents

        # Total the number of dependents.
        dependents = tax_unit("tax_unit_dependents", period)

        # Get their dependent exemption amount based on their filing status.
        return dependents * p.amount



from policyengine_us.model_api import *


class ar_personal_credit_dependent(Variable):
    value_type = float
    entity = Person
    label = "Arkansas personal tax credit dependent amount"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2021_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf#page=1"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=12"
    )
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        dependent = person("is_tax_unit_dependent", period)
        p = parameters(period).gov.states.ar.tax.income.credits.personal.amount
        return dependent * p.base



from policyengine_us.model_api import *


class il_mpe_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Illinois Medicaid Presumptive Eligibility eligible"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-120.66"
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        is_pregnant = person("is_pregnant", period.this_year)
        income_eligible = person("il_mpe_income_eligible", period)
        return is_pregnant & income_eligible



from policyengine_us.model_api import *


class de_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware EITC"
    unit = USD
    documentation = "Refundable or non-refundable Delaware EITC"
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf"
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        refundable_eitc = tax_unit("de_refundable_eitc", period)
        non_refundable_eitc = tax_unit("de_non_refundable_eitc", period)
        claims_refundable = tax_unit("de_claims_refundable_eitc", period)
        return where(claims_refundable, refundable_eitc, non_refundable_eitc)



from policyengine_us.model_api import *


class ar_retirement_or_disability_benefits_exemption_person(Variable):
    value_type = float
    entity = Person
    label = "Arkansas individual retirement or disability benefits exemption"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=13"
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.irs.income.exemption.traditional_distribution
        # Only head or spouse of the tax unit will have this exemption
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        disability_benefits_and_taxable_pensions = add(
            person, period, ["disability_benefits", "taxable_pension_income"]
        )
        # Filers over a certain age can deduct IRA distributions in addition to pension income
        ira_age_eligible = person("age", period) >= p.age_threshold
        age_eligible_ira_distributions = ira_age_eligible * person(
            "taxable_ira_distributions", period
        )
        return head_or_spouse * (
            disability_benefits_and_taxable_pensions
            + age_eligible_ira_distributions
        )



from policyengine_us.model_api import *


class il_tanf_countable_income_for_grant_calculation(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois Temporary Assistance for Needy Families (TANF) countable income for grant calculation"
    unit = USD
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.155"
    defined_for = StateCode.IL

    adds = [
        "il_tanf_countable_earned_income_for_grant_calculation",
        "il_tanf_countable_unearned_income",
    ]



from policyengine_us.model_api import *


class rent(Variable):
    value_type = float
    entity = Person
    label = "Rent"
    unit = USD
    definition_period = YEAR
    uprating = "gov.bls.cpi.cpi_u"

    def formula(person, period, parameters):
        pre_subsidy_rent = person("pre_subsidy_rent", period)
        total_pre_subsidy_rent = person.spm_unit.sum(pre_subsidy_rent)
        housing_subsidy = person.spm_unit("housing_assistance", period)
        total_rent = max_(total_pre_subsidy_rent - housing_subsidy, 0)
        rent_fraction = np.zeros_like(total_pre_subsidy_rent)
        mask = total_pre_subsidy_rent != 0
        rent_fraction[mask] = (
            pre_subsidy_rent[mask] / total_pre_subsidy_rent[mask]
        )
        return rent_fraction * total_rent



from policyengine_us.model_api import *


class az_aged_exemption(Variable):
    value_type = float
    entity = Person
    label = "Arizona aged exemption"
    unit = USD
    definition_period = YEAR
    defined_for = "az_aged_exemption_eligible_person"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.az.tax.income.exemptions

        age = person("age", period)
        return p.aged.calc(age)



from policyengine_us.model_api import *


class dc_tanf_meets_work_requirements(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Work requirement satisfied for DC Temporary Assistance for Needy Families (TANF)"
    definition_period = MONTH
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.19b",
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.19g",
    )
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        is_working = person("dc_tanf_is_working", period)
        work_requirement_exempt = person(
            "dc_tanf_work_requirement_exempt", period
        )
        meets_work_requirements = is_working | work_requirement_exempt
        return spm_unit.sum(~meets_work_requirements) == 0



from policyengine_us.model_api import *


class student_loan_interest_ald_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for the Student loan interest ALD"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/221"

    def formula(person, period, parameters):
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        filing_status = person.tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        is_aoc_eligible = person(
            "is_eligible_for_american_opportunity_credit", period
        )
        return head_or_spouse & ~separate & is_aoc_eligible



from policyengine_us.model_api import *


class mt_additions(Variable):
    value_type = float
    entity = Person
    label = "Montana additions to federal adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://rules.mt.gov/gateway/Subchapterhome.asp?scn=42%2E15%2E2",
        "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/05/Montana-Idividiual-Income-Tax-Return-Form-2-2022v6.2.pdf#page=4",
    )
    defined_for = StateCode.MT
    adds = "gov.states.mt.tax.income.additions.additions"



from policyengine_us.model_api import *


class snap_normal_allotment(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    documentation = "Normal SNAP benefit amount, equal to net income minus food contribution"
    label = "SNAP normal allotment"
    reference = "https://www.law.cornell.edu/uscode/text/7/2017#a"
    unit = USD
    defined_for = "is_snap_eligible"

    def formula(spm_unit, period, parameters):
        # Federal SNAP rules are defined in U.S.C Title 7, Chapter 51, which
        # also defines state powers to modify the rules.
        expected_contribution = spm_unit("snap_expected_contribution", period)
        max_allotment = spm_unit("snap_max_allotment", period)
        normal_allotment = max_allotment - expected_contribution
        min_allotment = spm_unit("snap_min_allotment", period)
        return max_(min_allotment, normal_allotment)



from policyengine_us.model_api import *


class ms_charitable_contributions_to_qualifying_foster_care_organizations(
    Variable
):
    value_type = float
    entity = TaxUnit
    label = "Charitable contributions to qualifying foster care organizations in Mississippi"
    definition_period = YEAR
    defined_for = StateCode.MS
    reference = "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100231.pdf#page=3"



from policyengine_us.model_api import *


class taxable_federal_pension_income(Variable):
    value_type = float
    entity = Person
    label = "Taxable federal pension income"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class spm_unit_broadband_subsidy(Variable):
    value_type = float
    entity = SPMUnit
    label = "SPM unit broadband subsidy"
    definition_period = YEAR
    unit = USD

    def formula(spm_unit, period, parameters):
        if parameters(period).gov.simulation.reported_broadband_subsidy:
            return spm_unit("spm_unit_broadband_subsidy_reported", period)
        return add(spm_unit, period, ["ebb", "acp"])



from policyengine_us.model_api import *


class pa_total_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "PA total taxable income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2021/2021_pa-40in.pdf#page=8"
    defined_for = StateCode.PA

    def formula(tax_unit, period, parameters):
        us_gross_income = add(tax_unit, period, ["irs_gross_income"])
        p = parameters(period).gov.states.pa.tax.income
        sources = p.nontaxable_income_sources
        pa_nontaxable_income = add(tax_unit, period, sources)
        return max_(0, us_gross_income - pa_nontaxable_income)



from policyengine_us.model_api import *


class vt_elderly_or_disabled_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Vermont elderly or disabled credit"
    documentation = "Schedule R credit for the elderly and the disabled"
    unit = USD
    reference = (
        "https://tax.vermont.gov/individuals/personal-income-tax/tax-credits"
    )
    defined_for = StateCode.VT
    # The Investment Tax Credit and Vermont Farm Income Averaging Credit are also subject to the match
    # these are currently not included

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.vt.tax.income.credits
        us_elderly_disabled_credit = tax_unit(
            "elderly_disabled_credit", period
        )
        return p.elderly_or_disabled * us_elderly_disabled_credit



from policyengine_us.model_api import *


class il_tanf_eligible_child(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Eligible child for Illinois Temporary Assistance for Needy Families (TANF) based on demographics"
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.60",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.dhs.tanf.age_threshold
        age = person("monthly_age", period)
        dependent = person("is_tax_unit_dependent", period)
        secondary_school_student = person("is_in_secondary_school", period)
        age_eligible = where(
            secondary_school_student,
            age <= p.student_dependent,
            age < p.minor_child,
        )

        return age_eligible & dependent



from policyengine_us.model_api import *


class pre_subsidy_care_expenses(Variable):
    value_type = float
    entity = Person
    label = "Pre subsidy care expenses"
    definition_period = MONTH
    unit = USD



from policyengine_us.model_api import *
from policyengine_core.simulations import *


class relative_capital_gains_mtr_change(Variable):
    value_type = float
    entity = Person
    label = "relative change in capital gains tax rate"
    unit = "/1"
    definition_period = YEAR

    def formula(person, period, parameters):
        simulation: Simulation = person.simulation
        baseline_branch = simulation.get_branch("baseline").get_branch(
            "baseline_cgr_measurement", clone_system=True
        )
        baseline_person = baseline_branch.populations["person"]
        baseline_branch.tax_benefit_system.neutralize_variable(
            "capital_gains_behavioral_response"
        )
        baseline_branch.set_input(
            "long_term_capital_gains_before_response",
            period,
            person("long_term_capital_gains_before_response", period),
        )
        baseline_mtr = baseline_person(
            "marginal_tax_rate_on_capital_gains", period
        )
        del simulation.branches["baseline"].branches[
            "baseline_cgr_measurement"
        ]

        measurement_branch = simulation.get_branch(
            "cgr_measurement", clone_system=True
        )
        measurement_branch.tax_benefit_system.neutralize_variable(
            "capital_gains_behavioral_response"
        )
        measurement_branch.set_input(
            "long_term_capital_gains_before_response",
            period,
            person("long_term_capital_gains_before_response", period),
        )
        measurement_person = measurement_branch.populations["person"]
        reform_mtr = measurement_person(
            "marginal_tax_rate_on_capital_gains", period
        )
        del simulation.branches["cgr_measurement"]

        # Handle zeros in tax rates to prevent log(0)
        min_rate = 0.001
        baseline_mtr_adj = np.maximum(baseline_mtr, min_rate)
        reform_mtr_adj = np.maximum(reform_mtr, min_rate)

        # Calculate log difference
        return np.log(reform_mtr_adj) - np.log(baseline_mtr_adj)


class capital_gains_elasticity(Variable):
    value_type = float
    entity = Person
    label = "elasticity of capital gains realizations"
    unit = "/1"
    definition_period = YEAR

    def formula(person, period, parameters):
        gov = parameters(period).gov
        return gov.simulation.capital_gains_responses.elasticity


class capital_gains_behavioral_response(Variable):
    value_type = float
    entity = Person
    label = "capital gains behavioral response"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        simulation = person.simulation
        if simulation.baseline is None:
            return 0

        if (
            parameters(
                period
            ).gov.simulation.capital_gains_responses.elasticity
            == 0
        ):
            return 0

        capital_gains = person(
            "long_term_capital_gains_before_response", period
        )
        tax_rate_change = person("relative_capital_gains_mtr_change", period)
        elasticity = person("capital_gains_elasticity", period)

        # Calculate response using log differences
        response_factor = np.exp(elasticity * tax_rate_change) - 1
        response = capital_gains * response_factor

        return response


class long_term_capital_gains_before_response(Variable):
    label = "capital gains before responses"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = USD
    uprating = "calibration.gov.irs.soi.long_term_capital_gains"


class adult_index_cg(Variable):
    value_type = int
    entity = Person
    label = "index of adult in household, ranked by capital gains"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (
            person.get_rank(
                person.household,
                -person("long_term_capital_gains_before_response", period),
                condition=person("is_adult", period),
            )
            + 1
        )


class marginal_tax_rate_on_capital_gains(Variable):
    label = "capital gains marginal tax rate"
    documentation = "Percent of marginal capital gains that do not increase household net income."
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "/1"

    def formula(person, period, parameters):
        mtr_values = np.zeros(person.count, dtype=np.float32)
        simulation = person.simulation
        DELTA = 1_000
        adult_index_values = person("adult_index_cg", period)
        for adult_index in [1, 2]:
            alt_simulation = simulation.get_branch(
                f"adult_{adult_index}_cg_rise"
            )
            mask = adult_index_values == adult_index
            for variable in simulation.tax_benefit_system.variables:
                variable_data = simulation.tax_benefit_system.variables[
                    variable
                ]
                if (
                    variable not in simulation.input_variables
                    and not variable_data.is_input_variable()
                ):
                    alt_simulation.delete_arrays(variable)
            alt_simulation.set_input(
                "capital_gains",
                period,
                person("capital_gains", period) + mask * DELTA,
            )
            alt_person = alt_simulation.person
            household_net_income = person.household(
                "household_net_income", period
            )
            household_net_income_higher_earnings = alt_person.household(
                "household_net_income", period
            )
            increase = (
                household_net_income_higher_earnings - household_net_income
            )
            mtr_values += where(mask, 1 - increase / DELTA, 0)

            del simulation.branches[f"adult_{adult_index}_cg_rise"]
        return mtr_values



from policyengine_us.model_api import *


class la_school_readiness_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana school readiness tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.revenue.louisiana.gov/IndividualIncomeTax/SchoolReadinessTaxCredit"
    defined_for = StateCode.LA

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(
            period
        ).gov.states.la.tax.income.credits.school_readiness
        # determine school readiness credit amount
        child_care_credit = add(
            tax_unit, period, ["la_refundable_cdcc", "la_non_refundable_cdcc"]
        )
        eligible_child = person(
            "la_school_readiness_credit_eligible_child", period
        )
        quality_rating = person(
            "la_quality_rating_of_child_care_facility", period
        )
        child_credit_percent = child_care_credit * p.rate.calc(quality_rating)
        # la_cdcc times the percentage
        amount = eligible_child * child_credit_percent

        return tax_unit.sum(amount)



import pandas as pd
import numpy as np
from policyengine_us.variables.household.demographic.geographic.county.county_enum import (
    County,
)
from pathlib import Path
from policyengine_core.tools.hugging_face import download_huggingface_dataset


def load_county_fips_dataset() -> pd.DataFrame:
    """
    Download the county FIPS dataset from Hugging Face and load it into a pandas DataFrame.
    If the dataset already exists in the 'data' folder and is the most recent version, this
    function will just load that into a pandas DataFrame.
    """

    DATA_FOLDER = Path("data")
    HUGGINGFACE_REPO = "policyengine/policyengine-us-data"
    COUNTY_FIPS_DATASET_FILENAME = "county_fips_2020.csv.gz"

    try:
        COUNTY_FIPS_RAW = download_huggingface_dataset(
            repo=HUGGINGFACE_REPO,
            repo_filename=COUNTY_FIPS_DATASET_FILENAME,
            version=None,
            local_dir=DATA_FOLDER,
        )

        # Read raw data into pandas dataframe; county FIPS MUST be defined as string,
        # else pandas reads as int and drops leading zeros
        COUNTY_FIPS_DATASET = pd.read_csv(
            COUNTY_FIPS_RAW,
            compression="gzip",
            dtype={"county_fips": str},
            encoding="utf-8",
        )

        return COUNTY_FIPS_DATASET

    except Exception as e:
        raise Exception(
            f"Error downloading {COUNTY_FIPS_DATASET_FILENAME} from {HUGGINGFACE_REPO}: {e}"
        )


def map_county_string_to_enum(
    county_name: "pd.Series[str]", state_code: "pd.Series[str]"
) -> "pd.Series[int]":
    """Helper function to map county name and state code to County enum value."""
    county_key = county_name.apply(
        lambda name: name.replace(" ", "_")
        .replace("-", "_")
        .replace(".", "")
        .replace("'", "_")
        .strip()
        .upper()
    )
    county_state = county_key.str.cat(state_code, sep="_")
    county_names = pd.Series(
        np.arange(len(County._member_names_)),
        index=County._member_names_,
    )
    return county_names[county_state]



from policyengine_us.model_api import *


class unemployment_compensation(Variable):
    value_type = float
    entity = Person
    label = "unemployment compensation"
    unit = USD
    documentation = "Income from unemployment compensation programs."
    definition_period = YEAR
    uprating = "calibration.gov.irs.soi.unemployment_compensation"



from policyengine_us.model_api import *
import pandas as pd


class SNAPRegion(Enum):
    CONTIGUOUS_US = "Contiguous US"
    AK_URBAN = "Alaska (urban)"
    AK_RURAL_1 = "Alaska (rural 1)"
    AK_RURAL_2 = "Alaska (rural 2)"
    GU = "Guam"
    HI = "Hawaii"
    VI = "Virgin Islands"


class snap_region(Variable):
    value_type = Enum
    possible_values = SNAPRegion
    default_value = SNAPRegion.CONTIGUOUS_US
    entity = Household
    definition_period = YEAR
    label = "SNAP region"

    def formula(household, period):
        state_group = household("state_group", period)
        mapped_values = (
            pd.Series(state_group.decode_to_str())
            .map(
                {
                    "AK": SNAPRegion.AK_URBAN,
                    **{
                        key: value
                        for key, value in SNAPRegion._member_map_.items()
                    },
                }
            )
            .values
        )
        return SNAPRegion.encode(mapped_values)



from policyengine_us.model_api import *


class ny_additions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York AGI additions"
    unit = USD
    documentation = "Additions to NY AGI over federal AGI."
    definition_period = YEAR
    dict(
        title="N.Y. Comp. Codes R. & Regs. tit. 20 § 112.2",
        href="https://casetext.com/regulation/new-york-codes-rules-and-regulations/title-20-department-of-taxation-and-finance/chapter-ii-income-taxes-and-estate-taxes/subchapter-a-new-york-state-personal-income-tax-under-article-22-of-the-tax-law/article-2-residents/part-112-new-york-adjusted-gross-income-of-a-resident-individual/section-1122-modifications-increasing-federal-adjusted-gross-income",
    )
    defined_for = StateCode.NY

    # No additions modeled in PolicyEngine US.



from policyengine_us.model_api import *


class is_medicare_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Person is eligible for Medicare"
    definition_period = YEAR
    documentation = (
        "CMS, Original Medicare (Part A and B) Eligibility and Enrollment"
        "https://www.cms.gov/medicare/enrollment-renewal/health-plans/original-part-a-b"
        "Above link includes the following text:"
        "  Part A coverage begins the month the individual turns age 65"
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.hhs.medicare.eligibility
        is_age_eligible = person("age", period) >= p.min_age
        gets_ssdi = person("social_security_disability", period) > 0
        months = person("months_receiving_social_security_disability", period)
        is_disability_eligible = gets_ssdi & (
            months >= p.min_months_receiving_social_security_disability
        )
        return is_age_eligible | is_disability_eligible



from policyengine_us.model_api import *


class meets_snap_parent_exception(Variable):
    value_type = bool
    entity = Person
    label = "Meets SNAP student parent exception"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/7/2015#e_5, "
        "https://www.law.cornell.edu/uscode/text/7/2015#e_8"
    )

    def formula(person, period, parameters):
        # Exception 5: Parent with responsibility for dependent child under 6,
        # Or child 6-11 when adequate child care is not available (not modeled)
        # Exception 8: Single parent enrolled full-time with responsibility
        # for dependent child under 12
        is_parent = person("is_tax_unit_head_or_spouse", period)
        spm_unit = person.spm_unit
        parent_count = spm_unit.sum(is_parent)

        # Check if there are children in the household under the age thresholds
        p = parameters(period).gov.usda.snap.student.child_age_limit
        household_member_ages = spm_unit.members("age", period)
        has_child_under_two_parent_limit = spm_unit.any(
            household_member_ages < p.two_parent
        )
        has_child_under_single_parent_limit = spm_unit.any(
            household_member_ages < p.single_parent
        )

        # Two-parent households need child under 6
        # Single parent households need child under 12, enrolled full-time
        is_full_time_student = person("is_full_time_college_student", period)
        parent_exception_requirement = where(
            parent_count > 1,
            has_child_under_two_parent_limit,
            has_child_under_single_parent_limit & is_full_time_student,
        )

        return is_parent & parent_exception_requirement



from policyengine_us.model_api import *


class nc_scca_entry_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "North Carolina entry income eligibility for Subsidized Child Care Assistance program"
    reference = "https://policies.ncdhhs.gov/wp-content/uploads/chapter-8-parental-fees-7.pdf#page=8"
    definition_period = MONTH
    defined_for = StateCode.NC

    def formula(spm_unit, period, parameters):
        family_total_income = spm_unit("nc_scca_countable_income", period)
        rounded_family_total_income = np.round(family_total_income, 2)

        fpg = spm_unit("spm_unit_fpg", period)

        rate = spm_unit("nc_scca_fpg_rate", period.this_year)

        allowed_max_income = np.round(fpg * rate, 2)

        return rounded_family_total_income < allowed_max_income



from policyengine_us.model_api import *


class monthly_hours_worked(Variable):
    value_type = float
    entity = Person
    label = "Average monthly hours worked"
    unit = "hour"
    definition_period = YEAR



from policyengine_us.model_api import *


class ar_gross_income_indiv(Variable):
    value_type = float
    entity = Person
    label = "Arkansas gross income when married filing separately"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/arkansas-code-of-1987/title-26-taxation/subtitle-5-state-taxes/chapter-51-income-taxes/subchapter-4-computation-of-tax-liability/section-26-51-404-gross-income-generally"
    defined_for = StateCode.AR

    adds = "gov.states.ar.tax.income.gross_income.sources.individual"



from policyengine_us.model_api import *


class mt_agi_joint(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana AGI for joint filers (tax unit level)"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT
    documentation = "Montana AGI calculated at tax unit level for joint filers, pooling income and subtractions before applying them"

    def formula(tax_unit, period, parameters):
        # Pool federal AGI at tax unit level
        agi = add(tax_unit, period, ["adjusted_gross_income_person"])

        # Pool Montana additions at tax unit level
        additions = add(tax_unit, period, ["mt_additions"])

        # Pool Montana subtractions at tax unit level
        # This allows spouse's subtractions to offset head's income
        subtractions = add(tax_unit, period, ["mt_subtractions"])

        # Apply pooled subtractions to pooled income
        reduced_agi = max_(agi + additions - subtractions, 0)

        # Montana taxable social security benefits can be either addition or subtraction
        # if mt_taxable_social_security - taxable_social_security > 0, then addition, else subtraction

        p = parameters(period).gov.states.mt.tax.income.social_security
        if p.applies:
            # 2023 and before: apply lines 21-24 adjustment for social security
            taxable_ss = add(tax_unit, period, ["taxable_social_security"])
            mt_taxable_ss = add(
                tax_unit, period, ["mt_taxable_social_security"]
            )
            adjusted_mt_ss_difference = mt_taxable_ss - taxable_ss
            tax_unit_mt_agi = max_(reduced_agi + adjusted_mt_ss_difference, 0)
        else:
            # 2024 and after: no longer apply the social security adjustment
            tax_unit_mt_agi = reduced_agi

        return tax_unit_mt_agi



from policyengine_us.model_api import *


class capped_electric_load_service_center_upgrade_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Capped electric load service center upgrade rebate"
    documentation = "Before total high efficiency electric home rebate cap"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        percent_covered = tax_unit(
            "high_efficiency_electric_home_rebate_percent_covered", period
        )
        expenditures = tax_unit(
            "electric_load_service_center_upgrade_expenditures", period
        )
        cap = parameters(
            period
        ).gov.doe.high_efficiency_electric_home_rebate.cap.electric_load_service_center_upgrade
        return min_(expenditures * percent_covered, cap)



from policyengine_us.model_api import *


class snap_net_income_fpg_ratio(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    label = "SNAP net income to FPL ratio"
    documentation = (
        "SNAP net income as a percentage of the federal poverty line"
    )
    unit = "/1"

    def formula(spm_unit, period):
        income = spm_unit("snap_net_income", period)
        fpg = spm_unit("snap_fpg", period)
        return income / fpg



from policyengine_us.model_api import *


class ma_eaedc_standard_assistance(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts EAEDC standard assistance"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://www.mass.gov/lists/emergency-aid-to-the-elderly-disabled-and-children-eaedc-grant-calculation"

    def formula(spm_unit, period, parameters):
        n = spm_unit("spm_unit_size", period)
        living_arrangement = spm_unit("ma_eaedc_living_arrangement", period)
        p = parameters(
            period
        ).gov.states.ma.dta.tcap.eaedc.standard_assistance.amount
        p1 = p.base[living_arrangement]
        pn = p.additional[living_arrangement]
        return p1 + pn * (n - 1)



from policyengine_us.model_api import *


class nm_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico child income tax credit"
    definition_period = YEAR
    unit = USD
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503818/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgAcHYQEoANMmylCEAIqJCuAJ7QA5BskRCYXAiUr1WnXoMgAynlIAhdQCUAogBknANQCCAOQDCTyVIwACNoUnZxcSA"
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        # Legal code refers to "adjusted gross income",
        # Tax form does not specify if federal or state AGI
        agi = tax_unit("adjusted_gross_income", period)
        # The law 7-2-18.34(J)(2) defines qualifying children as those from IRC 152(c).
        # IRC 152(c) refers to the EITC qualifying children.
        # https://www.law.cornell.edu/uscode/text/26/152#c
        children = tax_unit("eitc_child_count", period)
        p = parameters(period).gov.states.nm.tax.income.credits.ctc
        amount_per_child = p.amount.calc(agi, right=True)
        amount = amount_per_child * children
        # Halve the credit if married filing separately.
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        denominator = where(separate, 2, 1)
        return amount / denominator



from policyengine_us.model_api import *


class mt_elderly_homeowner_or_renter_credit_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for the Montana Elderly Homeowner/Renter Credit"
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.mt.tax.income.credits.elderly_homeowner_or_renter
        # If the filer is married and both spouses owned or rent the residence
        # Only one of them must meet the age requierments
        # reference: https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2022/12/Form-2-2022-Instructions.pdf#page=47
        greater_age_head_spouse = person.tax_unit(
            "greater_age_head_spouse", period
        )
        return greater_age_head_spouse >= p.age_threshold



from policyengine_us.model_api import *


class co_tanf_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Colorado TANF income eligible"
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(spm_unit, period, parameters):
        income = add(
            spm_unit,
            period,
            [
                "co_tanf_countable_earned_income_need",
                "co_tanf_countable_gross_unearned_income",
            ],
        )
        need_standard = spm_unit("co_tanf_need_standard", period)
        return income < need_standard



from policyengine_us.model_api import *


class pr_net_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Puerto Rico net taxable income"
    reference = "https://hacienda.pr.gov/sites/default/files/individuals_2024_rev._jul_12_24_9-30-24_informative.pdf#page=2"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.PR

    def formula(tax_unit, period, parameters):
        agi = tax_unit("pr_agi", period)
        exemptions = tax_unit("pr_exemptions", period)
        deductions = tax_unit("pr_deductions", period)
        return max_(0, agi - exemptions - deductions)



from policyengine_us.model_api import *


class la_general_relief_housing_subsidy_program_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the Los Angeles County General Relief Housing Subsidy based on the program eligibility"
    definition_period = YEAR
    # Person has to be a resident of LA County
    defined_for = "la_general_relief_eligible"
    reference = "https://dpss.lacounty.gov/en/cash/gr/housing.html"



from policyengine_us.model_api import *


class head_spouse_count(Variable):
    value_type = int
    entity = TaxUnit
    label = "Head and spouse count"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        return where(
            filing_status == filing_status.possible_values.JOINT, 2, 1
        )



from policyengine_us.model_api import *


class md_pension_subtraction_amount(Variable):
    value_type = float
    entity = Person
    label = "MD pension subtraction from AGI"
    unit = USD
    definition_period = YEAR
    reference = "https://www.marylandtaxes.gov/forms/21_forms/Resident_Booklet.pdf#page=13"
    defined_for = StateCode.MD

    def formula(person, period, parameters):
        p = parameters(period).gov.states.md.tax.income.agi.subtractions
        # determine pension subtraction eligiblity for each person
        dependent = person("is_tax_unit_dependent", period)
        min_age = p.pension.min_age
        elderly = person("age", period) >= min_age
        disabled = person("is_disabled", period)
        partner_is_disabled = person("has_disabled_spouse", period)
        eligible = ~dependent & (elderly | disabled | partner_is_disabled)
        # calculate pension subtraction amount for each person
        peninc = person("taxable_pension_income", period)
        socsec = person("social_security", period)
        amount = min_(peninc, max_(0, p.pension.max_amount - socsec))
        # return pension subtraction amount for each eligible person
        return eligible * amount



from policyengine_us.model_api import *


class pell_grant_uses_efc(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Pell Grant uses the expected family contribution"

    def formula(person, period, parameters):
        method = person.tax_unit("pell_grant_calculation_method", period)
        return method == method.possible_values.EFC



from policyengine_us.model_api import *


class adjusted_earnings(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Personal earned income adjusted for self-employment tax"
    unit = USD

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.ald.misc
        adjustment = (
            (1 - p.self_emp_tax_adj)
            * p.employer_share
            * person("self_employment_tax", period)
        )
        return max_(0, person("earned_income", period) - adjustment)



from policyengine_us.model_api import *


class sc_military_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina military retirement deduction"
    defined_for = StateCode.SC
    unit = USD
    reference = (
        "https://www.scstatehouse.gov/code/t12c006.php",  # SECTION 12-6-1171(A)
        "https://dor.sc.gov/forms-site/Forms/IITPacket_2021.pdf#page=17",
    )
    definition_period = YEAR

    adds = [
        "sc_military_deduction_indv",  # p4, p5
        "sc_military_deduction_survivors",  # p6
    ]



from policyengine_us.model_api import *


class additional_senior_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Senior deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.congress.gov/bill/119th-congress/house-bill/1/text"
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions.senior_deduction
        eligible_seniors = add(
            tax_unit,
            period,
            ["additional_senior_deduction_eligible_person"],
        )

        base_deduction = p.amount * eligible_seniors
        agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        joint = filing_status == filing_status.possible_values.JOINT
        phase_out_amount = where(
            joint,
            p.phase_out_rate.joint.calc(agi),
            p.phase_out_rate.other.calc(agi),
        )
        return max_(base_deduction - phase_out_amount, 0)



from policyengine_us.model_api import *


class co_care_worker_credit_eligible_care_worker(Variable):
    value_type = bool
    entity = Person
    label = "Eligible Care Worker for the Colorado Care Worker Tax Credit"
    defined_for = StateCode.CO
    definition_period = YEAR
    reference = "https://law.justia.com/codes/colorado/title-39/specific-taxes/income-tax/article-22/part-5/section-39-22-566/"



from policyengine_us.model_api import *


class il_smib(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Illinois SMIB Buy-In Program benefit"
    definition_period = MONTH
    documentation = (
        "Illinois Supplementary Medical Insurance Benefit (SMIB) Buy-In "
        "Program. The state pays Medicare Part B premiums for eligible "
        "individuals receiving AABD, TANF, SSI, or who qualify under the "
        "Medicare Savings Program (QMB, SLMB, QI)."
    )
    reference = (
        "https://www.ilga.gov/commission/jcar/admincode/089/089001200D00700R.html",
        "https://www.dhs.state.il.us/page.aspx?item=18685",
    )
    defined_for = StateCode.IL
    adds = ["il_smib_person"]



from policyengine_us.model_api import *


class eitc_maximum(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maximum EITC"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/32#a"
    unit = USD

    def formula(tax_unit, period, parameters):
        child_count = tax_unit("eitc_child_count", period)
        eitc = parameters(period).gov.irs.credits.eitc
        return eitc.max.calc(child_count)



from policyengine_us.model_api import *


class interest_income(Variable):
    value_type = float
    entity = Person
    label = "interest income"
    documentation = "Interest income from bonds, savings accounts, CDs, etc."
    unit = USD
    definition_period = YEAR
    adds = [
        "tax_exempt_interest_income",
        "taxable_interest_income",
    ]



from policyengine_us.model_api import *


class casualty_loss_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Casualty loss deduction"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        loss = add(tax_unit, period, ["casualty_loss"])
        p = parameters(period).gov.irs.deductions.itemized.casualty
        positive_agi = tax_unit("positive_agi", period)
        amount_over_floor = max_(0, loss - positive_agi * p.floor)
        return p.active * amount_over_floor



from policyengine_us.model_api import *


def create_medicare_and_investment_tax_increase() -> Reform:
    class additional_medicare_tax(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "Additional Medicare Tax"
        unit = USD
        documentation = (
            "Additional Medicare Tax from Form 8959 (included in payrolltax)"
        )

        def formula(tax_unit, period, parameters):
            amc = parameters(period).gov.irs.payroll.medicare.additional
            # Wage and self-employment income are taxed the same.
            ELEMENTS = [
                "irs_employment_income",
                "taxable_self_employment_income",
            ]
            wages_plus_se = add(tax_unit, period, ELEMENTS)
            exclusion = amc.exclusion[tax_unit("filing_status", period)]
            base = max_(0, wages_plus_se - exclusion)
            base_tax = amc.rate * base
            p_reform = parameters(
                period
            ).gov.contrib.biden.budget_2025.medicare
            add_excess = max_(wages_plus_se - p_reform.threshold, 0)
            add_tax = p_reform.rate * add_excess
            return base_tax + add_tax

    class net_investment_income_tax(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "Net Investment Income Tax"
        reference = "https://www.law.cornell.edu/uscode/text/26/1411"
        unit = USD

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.irs.investment.net_investment_income_tax
            threshold = p.threshold[tax_unit("filing_status", period)]
            agi = tax_unit("adjusted_gross_income", period)
            excess_agi = max_(0, agi - threshold)
            investment_income = tax_unit("net_investment_income", period)
            capped_investment_income = max_(0, investment_income)
            base = min_(
                capped_investment_income,
                excess_agi,
            )
            base_tax = p.rate * base
            p_reform = parameters(
                period
            ).gov.contrib.biden.budget_2025.net_investment_income
            add_excess_agi = max_(agi - p_reform.threshold, 0)
            lesser_of_excess_and_inv_income = min_(
                add_excess_agi, capped_investment_income
            )
            add_tax = p_reform.rate * lesser_of_excess_and_inv_income
            return base_tax + add_tax

    class reform(Reform):
        def apply(self):
            self.update_variable(additional_medicare_tax)
            self.update_variable(net_investment_income_tax)

    return reform


def create_medicare_and_investment_tax_increase_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_medicare_and_investment_tax_increase()

    p = parameters(period).gov.contrib.biden.budget_2025

    if (p.medicare.rate > 0) | (p.net_investment_income.rate > 0):
        return create_medicare_and_investment_tax_increase()
    else:
        return None


medicare_and_investment_tax_increase = (
    create_medicare_and_investment_tax_increase_reform(None, None, bypass=True)
)



from policyengine_us.model_api import *


class ma_tafdc_eligible_dependent(Variable):
    value_type = bool
    entity = Person
    label = "Eligible dependent for the Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC)"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-200"
    )
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        age = person("age", period)
        dependent = person("is_tax_unit_dependent", period)

        # Basic age eligibility: under standard dependent age threshold
        age_limit = person("ma_tafdc_age_limit", period)
        age_eligible = age < age_limit

        # Person must be a dependent and meet either age condition
        return dependent & age_eligible



from policyengine_us.model_api import *


class mn_wfc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota working family credit amount"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revisor.mn.gov/statutes/2021/cite/290.0671"
        "https://www.revisor.mn.gov/statutes/cite/290.0671"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.credits.cwfc
        # determine count of eligible dependents using EITC rules
        count = tax_unit("eitc_child_count", period)
        # determine pre-phaseout credit amount using EITC earnings
        earnings = tax_unit("filer_adjusted_earnings", period)
        capped_earn = min_(
            earnings,
            p.wfc.pre_cwfc_legislation.phase_in.earnings_maximum.calc(count),
        )
        amount = capped_earn * p.wfc.pre_cwfc_legislation.phase_in.rate.calc(
            count
        )
        # determine phaseout reduction
        agi = tax_unit("adjusted_gross_income", period)
        income = max_(earnings, agi)
        filing_status = tax_unit("filing_status", period)
        income_threshold = where(
            filing_status == filing_status.possible_values.JOINT,
            p.wfc.pre_cwfc_legislation.phase_out.threshold.joint.calc(count),
            p.wfc.pre_cwfc_legislation.phase_out.threshold.other.calc(count),
        )
        excess_income = max_(0, income - income_threshold)
        reduction = (
            excess_income
            * p.wfc.pre_cwfc_legislation.phase_out.rate.calc(count)
        )
        # determine credit amount after phaseout if eligible
        eligible = tax_unit("mn_wfc_eligible", period)
        return eligible * max_(0, amount - reduction)



from policyengine_us.model_api import *


class dc_ccsp_first_child_copay(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "DC Child Care Subsidy Program (CCSP) fist child copay amount"
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/Sliding%20Fee%20Scale.pdf"
    definition_period = MONTH
    defined_for = "dc_ccsp_is_youngest_child"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.dc.dhs.ccsp.copay.first_child
        countable_income = person.spm_unit("dc_ccsp_countable_income", period)
        fpg = person.spm_unit("spm_unit_fpg", period)
        income_to_fpg_ratio = countable_income / fpg

        is_full_time = person("dc_ccsp_is_full_time", period)
        copay_per_day = where(
            is_full_time,
            p.full_time.calc(income_to_fpg_ratio),
            p.part_time.calc(income_to_fpg_ratio),
        )
        attending_days_per_month = person(
            "dc_ccsp_attending_days_per_month", period
        )
        return copay_per_day * attending_days_per_month



from policyengine_us.model_api import *


class dc_pap_eligible_child(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Eligible child for DC Public Assistance Programs based on demographics"
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.18"
    )
    defined_for = "dc_tanf_immigration_status_eligible_person"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.dc.dhs.tanf.age_threshold
        age = person("monthly_age", period)
        dependent = person("is_tax_unit_dependent", period)
        secondary_school_student = person("is_in_secondary_school", period)
        age_limit = where(
            secondary_school_student, p.student_dependent, p.minor_child
        )

        return dependent & (age < age_limit)



from policyengine_us.model_api import *


class snap_unearned_income(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    label = "SNAP unearned income"
    documentation = "Unearned income for calculating the SNAP benefit"
    reference = "https://www.law.cornell.edu/cfr/text/7/273.9#b_2"
    unit = USD

    adds = "gov.usda.snap.income.sources.unearned"



from policyengine_us.model_api import *


class in_military_service_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana military service deduction"
    definition_period = YEAR
    unit = USD
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-2-4"  # (a)(1)
    defined_for = StateCode.IN

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = (
            parameters(period)
            .gov.states["in"]
            .tax.income.deductions.military_service
        )
        person_deduction = min_(
            person("military_service_income", period), p.max
        )
        return tax_unit.sum(person_deduction)



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_remove_head_of_household() -> Reform:
    class reform(Reform):
        def apply(self):
            self.neutralize_variable("head_of_household_eligible")

    return reform


def create_remove_head_of_household_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_remove_head_of_household()

    # Look ahead for the next five years
    p = parameters.gov.contrib.congress.romney.family_security_act
    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).remove_head_of_household:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_remove_head_of_household()
    else:
        return None


remove_head_of_household = create_remove_head_of_household_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class amt_income_less_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Alternative Minimum Tax Income less exemptions"
    unit = USD
    documentation = "Alternative Minimum Tax (AMT) income less exemptions"

    def formula(tax_unit, period, parameters):
        # Form 6251, Part I
        # Line 4
        amt_income = tax_unit("amt_income", period)
        # For filers subject to the kiddie tax, the deductions are not added back
        taxable_income = tax_unit("taxable_income", period)
        kiddie_tax_applies = tax_unit("amt_kiddie_tax_applies", period)
        applied_income = where(kiddie_tax_applies, taxable_income, amt_income)

        # Form 6251, Part II top
        # Line 5
        amt_exemption = tax_unit("amt_exemption", period)

        # Line 6
        return max_(0, applied_income - amt_exemption)



from policyengine_us.model_api import *


class taxsim_childcare(Variable):
    value_type = float
    entity = TaxUnit
    label = "Childcare"
    unit = USD
    definition_period = YEAR

    adds = ["tax_unit_childcare_expenses"]



from policyengine_us.model_api import *


class is_eligible_for_housing_assistance(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Is eligible for HUD voucher"
    documentation = "HUD housing assistance payment"
    definition_period = YEAR
    reference = "https://www.hud.gov/sites/dfiles/PIH/documents/HCV_Guidebook_Calculating_Rent_and_HAP_Payments.pdf"

    def formula(spm_unit, period, parameters):
        receives_housing_assistance = spm_unit(
            "receives_housing_assistance", period
        )
        income_level = spm_unit("hud_income_level", period)
        income_levels = income_level.possible_values
        is_income_eligible = (
            (income_level == income_levels.ESPECIALLY_LOW)
            | (income_level == income_levels.VERY_LOW)
            | (income_level == income_levels.LOW)
        )
        return receives_housing_assistance | is_income_eligible



from policyengine_us.model_api import *


class me_relief_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine Relief Rebate"
    defined_for = "me_relief_rebate_eligible"
    unit = USD
    definition_period = YEAR
    reference = "https://www.maine.gov/governor/mills/relief-checks"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.me.tax.income.credits.relief_rebate
        head_spouse_count = tax_unit("head_spouse_count", period)
        return head_spouse_count * p.amount



from policyengine_us.model_api import *


class spouse_earned(Variable):
    value_type = float
    entity = TaxUnit
    label = "Spouse's adjusted earnings"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        adjusted_earnings = person("adjusted_earnings", period)
        is_spouse = person("is_tax_unit_spouse", period)
        return tax_unit.sum(is_spouse * adjusted_earnings)



from policyengine_us.model_api import *


class ny_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY capped non-refundable tax credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        credits = parameters(
            period
        ).gov.states.ny.tax.income.credits.non_refundable
        income_tax = tax_unit("ny_income_tax_before_credits", period)
        total_credit_value = add(tax_unit, period, credits)
        return min_(income_tax, total_credit_value)



from policyengine_us.model_api import *


class ar_files_separately(Variable):
    value_type = bool
    entity = TaxUnit
    label = "married couple files separately on the Arkansas tax return"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2023_AR1000F_and_AR1000NR_Instructions.pdf"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2023_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
    )
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        itax_indiv = add(
            tax_unit,
            period,
            ["ar_income_tax_before_non_refundable_credits_indiv"],
        )
        itax_joint = add(
            tax_unit,
            period,
            ["ar_income_tax_before_non_refundable_credits_joint"],
        )
        return itax_indiv < itax_joint



from policyengine_us.model_api import *


class nc_use_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina use tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NC

    def formula(tax_unit, period, parameters):
        agi = tax_unit("adjusted_gross_income", period)
        p = parameters(period).gov.states.nc.tax.use_tax
        # Compute base amount, a dollar amount based on NC AGI.
        main_amount = p.base.calc(agi)
        # Switch to a percentage of income above the threshold.
        additional_amount = p.additional.calc(agi) * agi
        return main_amount + additional_amount



from policyengine_us.model_api import *


class marginal_tax_rate(Variable):
    label = "marginal tax rate"
    documentation = "Fraction of marginal income gains that do not increase household net income."
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "/1"

    def formula(person, period, parameters):
        netinc_base = person.household("household_net_income", period)
        delta = parameters(period).simulation.marginal_tax_rate_delta
        adult_count = parameters(period).simulation.marginal_tax_rate_adults
        sim = person.simulation
        mtr_values = np.zeros(person.count, dtype=np.float32)
        adult_indexes = person("adult_earnings_index", period)
        employment_income = person("employment_income", period)
        self_employment_income = person("self_employment_income", period)
        emp_self_emp_ratio = person("emp_self_emp_ratio", period)

        for adult_index in range(1, 1 + adult_count):
            alt_sim = sim.get_branch(f"mtr_for_adult_{adult_index}")
            for variable in sim.tax_benefit_system.variables:
                if (
                    variable not in sim.input_variables
                    or variable == "employment_income"
                ):
                    alt_sim.delete_arrays(variable)
            mask = adult_index == adult_indexes
            alt_sim.set_input(
                "employment_income",
                period,
                employment_income + mask * delta * emp_self_emp_ratio,
            )
            alt_sim.set_input(
                "self_employment_income",
                period,
                self_employment_income
                + mask * delta * (1 - emp_self_emp_ratio),
            )
            alt_person = alt_sim.person
            netinc_alt = alt_person.household("household_net_income", period)
            increase = netinc_alt - netinc_base
            mtr_values += where(mask, 1 - increase / delta, 0)
            del sim.branches[f"mtr_for_adult_{adult_index}"]
        return mtr_values


class adult_index(Variable):
    value_type = int
    entity = Person
    label = "Index of adult in household"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (
            person.get_rank(
                person.household,
                -person("age", period),
                condition=person("is_adult", period),
            )
            + 1
        )


class adult_earnings_index(Variable):
    value_type = int
    entity = Person
    label = "index of adult in household by earnings"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (
            person.get_rank(
                person.household,
                -person("market_income", period),
                condition=person("is_adult", period),
            )
            + 1
        )



from .end_child_poverty_act import (
    create_end_child_poverty_act_reform,
    end_child_poverty_act,
)
from .economic_dignity_for_all_agenda import (
    end_child_poverty_act as edaa_end_child_poverty_act,
    create_end_child_poverty_act_reform as create_edaa_end_child_poverty_act_reform,
)



from policyengine_us.model_api import *


class md_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD non-refundable credits"
    documentation = "Maryland non-refundable tax credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD

    adds = "gov.states.md.tax.income.credits.non_refundable"



from policyengine_us.model_api import *


class ma_eaedc_eligible_disabled_dependent_present(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Disabled dependent present for Massachusetts EAEDC"
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-340"
    )

    def formula(spm_unit, period, parameters):
        # Check if household has disabled dependents
        person = spm_unit.members
        is_disabled = person("is_disabled", period)
        is_dependent = person("is_tax_unit_dependent", period)
        disabled_dependent = is_disabled & is_dependent
        disabled_dependent_present = spm_unit.any(disabled_dependent)

        # If there are disabled dependents, check if at least one meets income eligibility
        p = parameters(period).gov.states.ma.dta.tcap.eaedc.income
        earned_income = person("ma_tcap_gross_earned_income", period)

        # A disabled dependent is income eligible if their income is below the limit
        disabled_dependent_income_eligible = disabled_dependent & (
            earned_income < p.disabled_limit
        )

        # The household is eligible if there's at least one income-eligible disabled dependent
        return disabled_dependent_present & spm_unit.any(
            disabled_dependent_income_eligible
        )



GROUP_ENTITY_NAMES = (
    "tax_unit",
    "spm_unit",
    "family",
    "household",
)


def test_variable_names():
    from policyengine_us import CountryTaxBenefitSystem
    from policyengine_us.model_api import STATES

    system = CountryTaxBenefitSystem()
    for instance in system.variables.values():
        variable = type(instance)
        if variable.__name__ in STATES:
            # State codes are allowed
            continue
        assert (
            variable.__name__.islower()
        ), f"{variable.__name__} is not all lowercase"



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def create_ri_exemption_reform() -> Reform:
    class ri_eligible_dependents_count(Variable):
        value_type = int
        entity = TaxUnit
        label = "Rhode Island eligible dependents count"
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.dependent_exemption

            person = tax_unit.members
            age = person("age", period)
            is_dependent = person("is_tax_unit_dependent", period)

            # Apply age limit if in effect
            if p.age_limit.in_effect:
                age_threshold = p.age_limit.threshold
                eligible_dependents = is_dependent & (age < age_threshold)
            else:
                eligible_dependents = is_dependent

            return tax_unit.sum(eligible_dependents)

    class ri_dependent_exemption_maximum(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island dependent exemption maximum before phaseout"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.dependent_exemption

            dependents_count = tax_unit("ri_eligible_dependents_count", period)
            return dependents_count * p.amount

    class ri_dependent_exemption_phaseout(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island dependent exemption phaseout amount"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.dependent_exemption

            filing_status = tax_unit("filing_status", period)
            agi = tax_unit("ri_agi", period)

            threshold = p.phaseout.threshold[filing_status]
            excess_income = max_(agi - threshold, 0)
            return excess_income * p.phaseout.rate

    class ri_dependent_exemption(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island dependent exemption"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):

            maximum = tax_unit("ri_dependent_exemption_maximum", period)
            phaseout = tax_unit("ri_dependent_exemption_phaseout", period)

            return max_(maximum - phaseout, 0)

    class ri_older_dependents_count(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island older dependents count"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            person = tax_unit.members
            is_dependent = person("is_tax_unit_dependent", period)
            total_dependents = tax_unit.sum(is_dependent)
            eligible_dependent_exemptions = tax_unit(
                "ri_eligible_dependents_count", period
            )
            return max_(0, total_dependents - eligible_dependent_exemptions)

    class ri_exemptions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island exemptions"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p_base = parameters(period).gov.states.ri.tax.income.exemption

            # Calculate personal exemptions base amount
            filing_status = tax_unit("filing_status", period)
            older_dependents = tax_unit("ri_older_dependents_count", period)
            personal_exemptions = (
                where(
                    filing_status == filing_status.possible_values.JOINT,
                    2,
                    1,
                )
                + older_dependents
            )
            personal_exemption_base = personal_exemptions * p_base.amount

            # Apply baseline phase-out to personal exemptions
            mod_agi = tax_unit("ri_agi", period)
            excess_agi = max_(0, mod_agi - p_base.reduction.start)
            increments = np.ceil(excess_agi / p_base.reduction.increment)
            reduction_rate = min_(p_base.reduction.rate * increments, 1)
            personal_exemption_amount = personal_exemption_base * (
                1 - reduction_rate
            )

            # Add dependent exemption (has its own phase-out logic)
            dependent_exemption_amount = tax_unit(
                "ri_dependent_exemption", period
            )

            return personal_exemption_amount + dependent_exemption_amount

    class reform(Reform):
        def apply(self):
            self.update_variable(ri_eligible_dependents_count)
            self.update_variable(ri_dependent_exemption_maximum)
            self.update_variable(ri_dependent_exemption_phaseout)
            self.update_variable(ri_dependent_exemption)
            self.update_variable(ri_exemptions)
            self.update_variable(ri_older_dependents_count)

    return reform


def create_ri_exemption_reform_fn(parameters, period, bypass: bool = False):
    if bypass:
        return create_ri_exemption_reform()

    p = parameters.gov.contrib.states.ri.dependent_exemption

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_ri_exemption_reform()
    else:
        return None


ri_exemption_reform = create_ri_exemption_reform_fn(None, None, bypass=True)



from policyengine_us.model_api import *


class il_base_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL base income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.IL

    adds = ["adjusted_gross_income", "il_base_income_additions"]
    subtracts = ["il_base_income_subtractions"]



from policyengine_us.model_api import *


class self_employment_income_behavioral_response(Variable):
    value_type = float
    entity = Person
    label = "self-employment income behavioral response"
    unit = USD
    definition_period = YEAR
    adds = ["labor_supply_behavioral_response"]
    subtracts = ["employment_income_behavioral_response"]



from policyengine_us.model_api import *


class me_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine main income tax (before credits and supplemental tax)"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("me_taxable_income", period)
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values

        p = parameters(period).gov.states.me.tax.income.main

        return select(
            [
                filing_status == status.SINGLE,
                filing_status == status.JOINT,
                filing_status == status.HEAD_OF_HOUSEHOLD,
                filing_status == status.SURVIVING_SPOUSE,
                filing_status == status.SEPARATE,
            ],
            [
                p.single.calc(taxable_income),
                p.joint.calc(taxable_income),
                p.head_of_household.calc(taxable_income),
                p.surviving_spouse.calc(taxable_income),
                p.separate.calc(taxable_income),
            ],
        )



from policyengine_us.model_api import *


class sc_senior_exemption_person(Variable):
    value_type = float
    entity = Person
    label = "South Carolina senior exemption for each person"
    unit = USD
    definition_period = YEAR
    reference = "https://dor.sc.gov/forms-site/Forms/SC1040_2022.pdf"
    defined_for = StateCode.SC

    def formula(person, period, parameters):
        # Get the SC senior exemptions part of the parameter tree
        p = parameters(period).gov.states.sc.tax.income.exemptions.senior

        # Get the individual filer's age.
        age = person("age", period)

        # Determine if head of household (filer) is eligible.
        age_eligible = age >= p.age_threshold

        head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        eligible = head_or_spouse & age_eligible
        # Get SC retirement income deduction and military retirement income deduction

        total_deductions = add(
            person,
            period,
            [
                "sc_retirement_deduction_indv",
                "sc_military_deduction_indv",
            ],
        )

        # Calculate senior exemption. The exemption can not be less than 0.
        reduced_max_amount = max_(p.amount - total_deductions, 0)

        return eligible * reduced_max_amount



from policyengine_us.model_api import *


class medical_expense_deduction(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Medical expense deduction"
    reference = "https://www.law.cornell.edu/uscode/text/26/213"
    unit = USD
    documentation = "Medical expenses deducted from taxable income."

    def formula(tax_unit, period, parameters):
        expense = add(tax_unit, period, ["medical_out_of_pocket_expenses"])
        p = parameters(period).gov.irs.deductions.itemized.medical
        medical_floor = p.floor * tax_unit("positive_agi", period)
        return max_(0, expense - medical_floor)



from policyengine_us.model_api import *


class md_total_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD total subtractions from AGI"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.md.tax.income.agi.subtractions
        total_subtractions = add(tax_unit, period, p.sources)
        # Prevent negative subtractions from acting as additions
        return max_(0, total_subtractions)



from policyengine_us.model_api import *


class monthly_age(Variable):
    value_type = float
    entity = Person
    label = "Monthly age"
    definition_period = MONTH

    def formula(person, period, parameters):
        # We divide annual variables by 12 when bringing them down to a month.
        # This reverses that to preserve the true age.
        yearly_age = person("age", period)
        return yearly_age * MONTHS_IN_YEAR



from policyengine_us.model_api import *


def create_end_child_poverty_act() -> Reform:
    class ecpa_adult_dependent_credit(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        unit = USD
        label = "End Child Poverty Act Adult Dependent Credit"

        def formula(tax_unit, period, parameters):
            person = tax_unit.members
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.end_child_poverty_act.adult_dependent_credit
            # Adult dependent credit.
            dependent = person("is_tax_unit_dependent", period)
            adult = person("age", period) >= p.min_age
            return p.amount * tax_unit.sum(adult & dependent)

    class ecpa_filer_credit(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        unit = USD
        label = "End Child Poverty Act Filer Credit"
        reference = "https://tlaib.house.gov/sites/tlaib.house.gov/files/EndChildPovertyAct.pdf"

        def formula(tax_unit, period, parameters):
            # Filer credit.
            # Define eligibility based on age.
            age_head = tax_unit("age_head", period)
            age_spouse = tax_unit("age_spouse", period)
            p = parameters(
                period
            ).gov.contrib.congress.tlaib.end_child_poverty_act.filer_credit

            head_qualifies = (age_head >= p.eligibility.min_age) & (
                age_head <= p.eligibility.max_age
            )
            spouse_qualifies = (age_spouse >= p.eligibility.min_age) & (
                age_spouse <= p.eligibility.max_age
            )
            filer_credit_eligible = head_qualifies | spouse_qualifies
            # Get maximum amount.
            filing_status = tax_unit("filing_status", period)
            max_filer_credit = p.amount[filing_status]
            # Phase out.
            agi = tax_unit("adjusted_gross_income", period)
            phase_out_start = p.phase_out.start[filing_status]
            excess = max_(agi - phase_out_start, 0)
            reduction = excess * p.phase_out.rate
            # Compute final amount.
            return filer_credit_eligible * max_(
                max_filer_credit - reduction, 0
            )

    class ecpa_child_benefit(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "End Child Poverty Act Child Benefit"
        unit = USD

        def formula(tax_unit, period, parameters):
            person = tax_unit.members
            dependent = person("is_tax_unit_dependent", period)
            age = person("age", period)
            p_ecpa = parameters(
                period
            ).gov.contrib.congress.tlaib.end_child_poverty_act.child_benefit
            age_eligible = age < p_ecpa.age_limit
            eligible_dependent = dependent & age_eligible
            total_dependents = tax_unit.sum(eligible_dependent)
            state_group = tax_unit.household("state_group_str", period)
            p_fpg = parameters(period).gov.hhs.fpg
            amount = p_fpg.additional_person[state_group]

            return total_dependents * amount

    class household_benefits(Variable):
        value_type = float
        entity = Household
        label = "benefits"
        unit = USD
        definition_period = YEAR
        adds = [
            "social_security",
            "ssi",
            "snap",
            "wic",
            "free_school_meals",
            "reduced_price_school_meals",
            "spm_unit_broadband_subsidy",
            "tanf",
            "high_efficiency_electric_home_rebate",
            "residential_efficiency_electrification_rebate",
            "unemployment_compensation",
            # Contributed.
            "basic_income",
            "spm_unit_capped_housing_subsidy",
            "household_state_benefits",
            "ecpa_child_benefit",
        ]

    class spm_unit_benefits(Variable):
        value_type = float
        entity = SPMUnit
        label = "Benefits"
        definition_period = YEAR
        unit = USD

        def formula(spm_unit, period, parameters):
            BENEFITS = [
                "social_security",
                "ssi",
                "ma_state_supplement",  # Massachusetts benefits
                # California programs.
                "ca_cvrp",  # California Clean Vehicle Rebate Project.
                # Colorado programs.
                "co_ccap_subsidy",
                "co_state_supplement",
                "co_oap",
                "snap",
                "wic",
                "free_school_meals",
                "reduced_price_school_meals",
                "spm_unit_broadband_subsidy",
                "spm_unit_energy_subsidy",
                "tanf",
                "high_efficiency_electric_home_rebate",
                "residential_efficiency_electrification_rebate",
                "unemployment_compensation",
                # Contributed.
                "basic_income",
                "ny_drive_clean_rebate",
                "ecpa_child_benefit",
            ]
            if parameters(period).gov.contrib.ubi_center.flat_tax.deduct_ptc:
                BENEFITS.append("aca_ptc")
            if not parameters(period).gov.hud.abolition:
                BENEFITS.append("spm_unit_capped_housing_subsidy")
            return add(spm_unit, period, BENEFITS)

    class income_tax_refundable_credits(Variable):
        value_type = float
        entity = TaxUnit
        definition_period = YEAR
        label = "federal refundable income tax credits"
        unit = USD

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.irs.credits
            previous_credits = add(tax_unit, period, p.refundable)
            filer_credit = tax_unit("ecpa_filer_credit", period)
            adult_dependent_credit = tax_unit(
                "ecpa_adult_dependent_credit", period
            )
            return filer_credit + adult_dependent_credit + previous_credits

    class reform(Reform):
        def apply(self):
            self.update_variable(ecpa_adult_dependent_credit)
            self.update_variable(ecpa_filer_credit)
            self.update_variable(income_tax_refundable_credits)
            self.update_variable(ecpa_child_benefit)
            self.update_variable(household_benefits)
            self.update_variable(spm_unit_benefits)
            self.neutralize_variable("eitc")
            self.neutralize_variable("ctc")

    return reform


def create_end_child_poverty_act_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_end_child_poverty_act()

    p = parameters(period).gov.contrib.congress.tlaib.end_child_poverty_act

    if p.in_effect:
        return create_end_child_poverty_act()
    else:
        return None


end_child_poverty_act = create_end_child_poverty_act_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class il_hbwd_employment_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Illinois Health Benefits for Workers with Disabilities employment eligible"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-120.510",
        "https://hfs.illinois.gov/medicalprograms/hbwd/eligibility.html",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        # Per § 120.510(l)(1), employment is verified by current payment
        # under FICA or IMRF. We use earned income as a proxy for
        # employment with FICA contributions.
        #
        # Not currently modeled:
        # - § 120.510(l)(2)(A): Future employment exception (60 days)
        # - § 120.510(l)(2)(B): Medical inability grace period (90 days)
        # - § 120.510(l)(2)(C): Job loss grace period (30 days)
        earned_income = person("il_hbwd_gross_earned_income", period)
        return earned_income > 0



from policyengine_us.model_api import *


class mt_elderly_homeowner_or_renter_credit_gross_household_income(Variable):
    value_type = float
    entity = Person
    label = "Montana gross household income for the elderly homeowner/renter credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    adds = "gov.states.mt.tax.income.credits.elderly_homeowner_or_renter.gross_income_sources"



from policyengine_us.model_api import *


class is_parent_of_filer_or_spouse(Variable):
    value_type = bool
    entity = Person
    label = "Is a parent of the filer or spouse"
    definition_period = YEAR
